import numpy as np
import itertools

class _ConfusionMatrixMixin(object):
  """
  Mixin for confusion matrix functionalities that are attached to `libraries.libraries.logger.Logger`.

  This mixin cannot be instantiated because it is built just to provide some additional
  functionalities for `libraries.libraries.logger.Logger`

  In this mixin we can use any attribute/method of the Logger.

  * Obs: This mixin uses also attributes/methods of `_MatplotlibMixin`:
    - self.add_copyright_to_plot
    - self.save_plot
  """

  def __init__(self):
    super(_ConfusionMatrixMixin, self).__init__()

    try:
      from .matplotlib_mixin import _MatplotlibMixin
    except ModuleNotFoundError:
      raise ModuleNotFoundError("Cannot use _ConfusionMatrixMixin without having _MatplotlibMixin")

    return

  def plot_confusion_matrix(self,
                            cm, classes=["0", "1"],
                            normalize=True,
                            title='Confusion matrix',
                            no_save=False,
                            cmap=None,
                            figsize=None,
                            titlefontsize=None,
                            axesfontsize=None,
                            ):
    """
    This function prints and plots the confusion matrix.
    Normalization can be applied by setting `normalize=True`.

    Parameters
    ----------
    cm : ndarray
      confusion matrix from sklearn.metric.confusion_matrix or identic format.
    classes : list of strings, optional
      label for each class. The default is ["0", "1"].
    normalize : bool, optional
      normalize cells (lines will add to 100%). The default is True.
    title : string, optional
      Title of plot. The default is 'Confusion matrix'.
    no_save : bool, optional
      skip save. The default is False.
    cmap : colormap, optional
      pyplot colormap. The default is None.
    figsize : int/tuple, optional
      int or tuple for figsize. The default is None.
    titlefontsize, axesfontsize: int, optional
      font size for title and axes


    Returns
    -------
    None.

    """

    if normalize:
      cm = cm.astype('float') / cm.sum(axis=1)[:, np.newaxis]
      s_title = title + " [norm]"
    else:
      s_title = title + " [raw]"

    if len(classes) != cm.shape[0]:
      classes = [str(x) for x in range(cm.shape[0])]

    if len(classes) < 4:
      MIN_FIG_SIZE = 9
    else:
      MIN_FIG_SIZE = 13

    import matplotlib.pyplot as plt
    from textwrap import wrap

    if figsize is None:
      figsize = (MIN_FIG_SIZE, MIN_FIG_SIZE)
    else:
      if type(figsize) == int:
        figsize = (figsize, figsize)
      if figsize[0] < MIN_FIG_SIZE or figsize[1] < MIN_FIG_SIZE:
        raise ValueError("Figsize must be at least {} for a decent plot".format(
          MIN_FIG_SIZE))

    titlefontsize = 20 if titlefontsize is None else titlefontsize
    axesfontsize = 16 if axesfontsize is None else axesfontsize

    if cmap is None:
      cmap = plt.cm.Blues
    fig = plt.figure(figsize=figsize)
    plt.imshow(cm, cmap=cmap, interpolation='nearest')
    plt.colorbar()
    plt.ylabel(
      'Reality (each line add to total 100%)' if normalize else 'Reality (count)',
      labelpad=25,
      fontsize=axesfontsize,
    )
    plt.xlabel(
      'Predicted ' + ('percentages' if normalize else 'counts') + ' for each class',
      labelpad=25,
      fontsize=axesfontsize,
    )

    tick_marks = np.arange(start=-1, stop=cm.shape[0] + 1)
    lbs = [' '] + classes + [' ']
    plt.xticks(ticks=tick_marks, labels=lbs, rotation=45)
    plt.yticks(ticks=tick_marks, labels=lbs)

    plt.title(
      "\n".join(wrap(s_title, 30)),
      y=1.05,
      loc='center',
      # family="Times New Roman",
      fontsize=titlefontsize,
      fontweight='bold',
      color='navy',
      wrap=True,
    )

    thresh = cm.max() / 2.
    for i, j in itertools.product(range(cm.shape[0]), range(cm.shape[1])):
      num = "{:.2f}".format(cm[i, j]) if normalize else cm[i, j]
      plt.text(j, i, num,
               horizontalalignment="center",
               color="white" if cm[i, j] > thresh else "black")

    self.add_copyright_to_plot(plt, push_right=True)

    # fig.tight_layout()
    # offset = max(0,(3-nr_series) * 0.05)
    # fig.subplots_adjust(top=0.95 - offset)
    if not no_save:
      _lbl = s_title.replace(" ", "_").lower()
      self.save_plot(plt, label=_lbl, include_prefix=False, just_label=True)
    plt.show()
    return

  def log_confusion_matrix(self,
                           cm=None,
                           y_true=None, y_pred=None,
                           classes=None,
                           title='Confusion Matrix',
                           normalize=None,
                           hide_zeroes=False,
                           hide_diagonal=False,
                           hide_threshold=None,
                           ):
    """
    pretty print for confusion matrixes. You must pass either a CM or y_true, y_pred
    """
    from sklearn.metrics import confusion_matrix

    if cm is None and (y_true is None or y_pred is None):
      self.raise_error("You must either have `cm` `y_true`/`y_pred`")

    if cm is not None and classes is None:
      self.raise_error("If you provide `cm` please also provide labels")

    if y_true is not None and y_pred is not None:
      classes = [str(x) for x in np.unique(y_true)]
      cm = confusion_matrix(y_true=y_true, y_pred=y_pred)

    labels = classes
    columnwidth = max([len(x) for x in labels] + [8])  # 5 is value length
    empty_cell = " " * columnwidth
    full_str = "         " + empty_cell + "Preds\n"
    n_classes = cm.shape[0]

    if len(labels) != n_classes:
      self.raise_error("Confusion matrix classes differ from number of labels")
    if n_classes > 2:
      max_lab = max([len(x) for x in labels])
      s1 = "  {:>" + str(max_lab) + "}"
      self.P("{} class breakdown:".format(title))
      self.P((s1 + " {:>7} {:>7} [{:>7} {:>7} {:>7}]").format(
        " " * max_lab,
        "TP",
        "GT",
        "REC",
        "PRE",
        "F1"
      ))
      f1scores = []
      for i, k in enumerate(labels):
        tp = cm[i, i]
        fp = cm[:, i].sum() - tp
        gt = cm[i].sum()
        rec = tp / gt
        pre = tp / (tp + fp)
        f1 = 2 * (pre * rec) / (pre + rec)
        f1scores.append(f1)
        self.P((s1 + ":{:>7}/{:>7} [{:>6.1f}% {:>6.1f}% {:>6.1f}%]").format(
          k, tp, gt, rec * 100, pre * 100, f1 * 100))
      f1macro = np.mean(f1scores)
      self.P("  Macro F1: {:.2f}%".format(f1macro * 100))
    if normalize is None:
      if cm.shape[0] > 2:
        normalize = True
      else:
        normalize = False

    if normalize:
      cm = cm.astype('float') / cm.sum(axis=1)[:, np.newaxis]
    full_str += "    " + empty_cell + " "
    for label in labels:
      full_str += "%{0}s".format(columnwidth) % label + " "
    full_str += "\n"
    # Print rows
    for i, label1 in enumerate(labels):
      if i == 0:
        full_str += "GT  %{0}s".format(columnwidth) % label1 + " "
      else:
        full_str += "    %{0}s".format(columnwidth) % label1 + " "
      for j in range(len(labels)):
        num = round(cm[i, j], 2) if normalize else cm[i, j]
        # cell = '{num:{fill}{width}}'.format(num=num, fill=' ', width=columnwidth)
        cell = ('{:>' + str(columnwidth) + '.2f}').format(num) if normalize else '{num:{fill}{width}}'.format(num=num,
                                                                                                              fill=' ',
                                                                                                              width=columnwidth)
        # "%{0}.0f".format(columnwidth) % cm[i, j]
        if hide_zeroes:
          cell = cell if float(cm[i, j]) != 0 else empty_cell
        if hide_diagonal:
          cell = cell if i != j else empty_cell
        if hide_threshold:
          cell = cell if cm[i, j] > hide_threshold else empty_cell
        full_str += cell + " "
      full_str += "\n"
    self.P("{}:\n{}".format(title, full_str))
    return



import numpy as np
from copy import deepcopy

from naeural_core import Logger
from naeural_core.bc import DefaultBlockEngine, BCct

if __name__ == '__main__':
  
  log = Logger('JSN', base_folder='.', app_folder='_local_cache')
  
  bc_engine = DefaultBlockEngine(
      log=log,
      name="local",
      config= {
        "PEM_FILE"        : "e2.pem",
        "PASSWORD"        : None,      
        "PEM_LOCATION"    : "data"
      },
    )  
  
  dct = {
  'a' : float('nan'), 
  'b' : [
    {
      'ba' : float('nan'), 
      'bc': float('-inf'),
      'bi': float('inf'),
      'bn' : 100,
    },
    {
      'b1' : {'b1a' : float('nan')}
    }
    ], 
  'c': float('-inf'), 
  'd': np.array([1,2,3]), 
  'e': np.array([1], dtype='int64')[0]
  }  
  
  d2 = log.replace_nan(dct)
  log.P(log.safe_json_dumps(dct))
  log.P(log.safe_json_dumps(d2))
  
  dct2 = deepcopy(dct)
  
  s1 = bc_engine.sign(dct)
  assert dct[BCct.SIGN] == s1
  s2 = bc_engine.sign(dct2, replace_nan=True)
  assert dct2[BCct.SIGN] == s2
  s3 = bc_engine.sign(d2)
  assert d2[BCct.SIGN] == s3
  
  assert dct2[BCct.HASH] == d2[BCct.HASH]
  assert dct2[BCct.SENDER] == d2[BCct.SENDER]
  log.P(s1)
  log.P(s2)
  log.P(s3)
  


from naeural_core.serving.default_inference.th_yf8s import _CONFIG as Y8S_CONFIG
from naeural_core.serving.default_inference.th_yf_mspn import ThYfMspn as BaseServingProcess

_CONFIG = {
  **BaseServingProcess.CONFIG,
  **Y8S_CONFIG,

  'MAX_BATCH_SECOND_STAGE': 10,

  'COVERED_SERVERS': ['th_yf8s'],

  'VALIDATION_RULES': {
    **BaseServingProcess.CONFIG['VALIDATION_RULES'],
  },
}

DEBUG_COVERED_SERVERS = False


class ThYfsMspn(BaseServingProcess):
  CONFIG = _CONFIG

  def __init__(self, **kwargs):
    super(ThYfsMspn, self).__init__(**kwargs)

    return



#global dependencies
import pandas as pd

from datetime import datetime as dt
from dateutil.relativedelta import relativedelta

class _IntervalAggregationMixin(object):

  def __init__(self):
    self._aggregation_buffer = []
    self._aggregation_times = []
    super(_IntervalAggregationMixin, self).__init__()
    return

  @property
  def cfg_interval_aggregation_seconds(self):
    return self._instance_config.get('INTERVAL_AGGREGATION_SECONDS', 0)

  def aggregation_add_to_buffer(self, obj):
    self._aggregation_buffer.append(obj)
    full_date_fmt = '%Y-%m-%d %H:%M:%S'
    relative_time_fmt = '%H:%M:%S'

    t = None
    is_last_data = False
    fmt = full_date_fmt
    if hasattr(self, 'is_data_limited'):
      if self.is_data_limited_and_has_frame:
        if self.is_last_data:
          is_last_data = True
        t = dt.strptime(self.limited_data_crt_time, '{}.%f'.format(relative_time_fmt))
        fmt = relative_time_fmt

    if t is None:
      t = dt.now()

    self._aggregation_times.append(t)
    intervals = []

    df = pd.DataFrame(
      {
        'ts'  : self._aggregation_times,
        'idx' : list(range(len(self._aggregation_times)))
      }
    )

    interval_seconds = self.cfg_interval_aggregation_seconds
    if interval_seconds == 0:
      freq = '1ms'
    else:
      freq = '{}s'.format(interval_seconds)

    groups = df.groupby(pd.Grouper(key='ts', freq=freq))

    last_idx = -1
    for i, (_time, df_grp) in enumerate(groups):
      start_time = _time
      end_time = _time + relativedelta(seconds=interval_seconds)
      idx = df_grp['idx'].values

      # at tick, the interest is for the last added object, whereas for specific intervals
      # the interest is for all the past intervals except the current one which is not finished
      if i == len(groups)-1 and interval_seconds > 0 and not is_last_data:
        break

      objects = []
      for j in idx:
        objects.append(self._aggregation_buffer[j])
        last_idx = j

      intervals.append({
        'START' : start_time.strftime(fmt),
        'END'   : end_time.strftime(fmt),
        'OBJECTS' : objects
      })
    #endfor

    self._aggregation_buffer = self._aggregation_buffer[last_idx+1:]
    self._aggregation_times = self._aggregation_times[last_idx+1:]

    if len(intervals) == 0:
      last_interval = None
      null_intervals = []
    else:
      last_interval = intervals[0]
      null_intervals = intervals[1:]

    ### sanity check that null intervals are really null :)
    for interval in null_intervals:
      objects = interval.pop('OBJECTS')
      if len(objects) > 0:
        self.P("WARNING! There is a coding problem in `aggregation_add_to_buffer`", color='e')
    #endfor

    return last_interval, null_intervals

# if __name__ == '__main__':
#
#   from time import sleep
#
#   class Base(object):
#     def __init__(self, **kwargs):
#       self._instance_config = {
#         'INTERVAL_AGGREGATION_SECONDS' : 0
#       }
#       return
#
#
#   class P(_IntervalAggregationMixin, Base):
#     def __init__(self, **kwargs):
#       super(P, self).__init__(**kwargs)
#       return
#
#
#   p = P()
#   for step in range(10):
#     _last_interval, _null_intervals = p.aggregation_add_to_buffer({'INFERENCE' : step+1})
#     print(_last_interval)
#     print(len(_null_intervals))
#     sleep(1)
#     # if step == 10:
#     #   sleep(20)
#



import pandas as pd
import datetime
import numpy as np

from naeural_core.core_logging import SBLogger


if __name__ == "__main__":
  log=SBLogger()

  NUMBER_OF_DATAPOINTS = 5000
  TIMEDELTA_MINUTES = 60
  ANOMALY_RATE = 0.01

  base = datetime.datetime.today()
  timestamps = [base - datetime.timedelta(minutes=TIMEDELTA_MINUTES * x) for x in range(NUMBER_OF_DATAPOINTS)]
  values = [
    np.random.normal(loc=0, scale=10) if np.random.uniform(0, 1) > ANOMALY_RATE
      else np.random.uniform(100,200)
    for _ in range(NUMBER_OF_DATAPOINTS)
  ]

  df = pd.DataFrame({'timestamps':timestamps, 'vals':values})
  log.save_dataframe(
    df=df,
    fn='tutorial_ts.csv'
  )




"""
{
	"NAME" : "auto_train_full_process",
	"TYPE" : "void",
	"PLUGINS" : [
		{
			"SIGNATURE" : "ai4e_end_to_end_training",
			"INSTANCES" : [
				{
					"INSTANCE_ID" : "alabala",  # job identifier
					"OBJECTIVE_NAME" : "weapons",  # job name
					"GENERAL_DETECTOR_OBJECT_TYPE" : ["person"],  # what will be cropped
					"GENERAL_DETECTOR_AI_ENGINE" : "general_detector",  # the engine that will be used for cropping
          "CLASSES": {  # the classes that will be trained and their descriptions
            "weapon": "Weapon class",
            "no_weapon": "No weapon class"
          },
          "DESCRIPTION": "Classify people based on if they have a weapon or not",  # job description
          'REWARDS': {  # rewards for the job
            "BUDGET": 1000,  # the budget for the job
            "SEATS": 5  # the number of seats for the job
          },
          'DATASET': {
            'MAX_SIZE': 1000,  # the maximum size of the dataset
          },
          'CREATION_DATE': "2022-08-10 14:55",  # the creation date of the job
					"DATA" : {  # the data sources
						"SOURCES" : [
							{
								"NAME" : "terasa",  # the name of the source
								"TYPE" : "VideoStream",  # the type of the source
								"URL"  : "__URL__",  # the url of the source
                # below are advanced parameters
								"LIVE_FEED" : true,  # if the source will be processed live or frame by frame
								"CAP_RESOLUTION" : 0.5,  # how many frames per second will be read
								"RECONNECTABLE" : true,  # if the source will be reconnected if it disconnects
								"STREAM_CONFIG_METADATA" : {
									"INTERVALS" : {  # this can be used if we only want to process the source at certain intervals
										"ziua" : ["10:00", "17:00"],
										"noaptea" : ["21:00", "23:59"]
									}
								}
							},

							{
								"NAME" : "openspace_est",
								"TYPE" : "VideoStream",
								"URL"  : "__URL__",
								"LIVE_FEED" : true,
								"RECONNECTABLE" : true,
								"CAP_RESOLUTION" : 0.5,
								"STREAM_CONFIG_METADATA" : {
									"INTERVALS" : {
										"ziua" : ["10:00", "17:00"],
										"noaptea" : ["21:00", "23:59"]
									}
								}
							}
						],

						"CROP_PLUGIN_PARAMS" : {  # parameters for the cropping plugin
							"REPORT_PERIOD" : 60,  # how often the plugin will report the progress
							"ALIVE_UNTIL" : "2022-08-10 14:55"  # until when the plugin will be alive(in case of not stopping it manually or in case we don't gather enough data)
						},

						"CLOUD_PATH" : "DATASETS/"  # the path where the dataset will be stored
					},

          "START_TRAINING": true,  # if the training should start automatically
					"TRAINING" : {  # the training parameters
						"BOX_ID" : "hidra-training",  # the id of the box where the training will be done
						"DEVICE_LOAD_DATA" : "cuda:3",  # the device where the data will be loaded
						"DEVICE_TRAINING"  : "cuda:3",  # the device where the training will be done
						"TRAINING_PIPELINE_SIGNATURE" : "weapons",  # the signature of the training pipeline
						"GRID_SEARCH" : {},  # the grid search parameters (if none are provided the defaults will be used)
						"BATCH_SIZE" : 64,  # the batch size
						"EPOCHS" : 4  # the number of epochs
					},

					"AUTO_DEPLOY" : {  # the auto deploy parameters
						"STREAMS" : [  # the streams that will be deployed
							{  # For details on the stream configuration see the stream configuration section
								"NAME" : "terasa",
								"TYPE" : "VideoStream",
								"URL"  : "__URL__",
								"LIVE_FEED" : true,
								"CAP_RESOLUTION" : 20,
								"RECONNECTABLE" : true
							}
						]
					}
				}
			]
		}
	]
}
"""

from naeural_core.business.base import BasePluginExecutor as BasePlugin

__VER__ = '0.1.0.0'

DEFAULT_DATA_CONFIG = {
  'SOURCES': [],
  "CROP_PLUGIN_PARAMS": {
    "REPORT_PERIOD": 5
  },
  "COLLECT_UNTIL": None,
  "CLOUD_PATH": "DATASETS/",
  "TRAIN_SIZE": 0.8,
}

DEFAULT_TRAIN_CONFIG = {
  'BOX_ID': None,
  'DEVICE_LOAD_DATA': None,  # None for no preloading data, string for device otherwise
  'DEVICE_TRAINING': 'cuda:0',
  'TRAINING_PIPELINE_SIGNATURE': 'custom',
  'MODEL_ARCHITECTURE': 'BASIC_CLASSIFIER',
  'GRID_SEARCH': {},
  'BATCH_SIZE': 8,
  'EPOCHS': 4,
}

_CONFIG = {
  **BasePlugin.CONFIG,
  'OBJECTIVE_NAME': None,
  'GENERAL_DETECTOR_OBJECT_TYPE': ['person'],
  "CLASSES": {},
  "DESCRIPTION": "",
  'DATA': {},
  'REWARDS': {},
  'DATASET': {},
  'CREATION_DATE': None,
  'TRAINING_REWARDS': {},

  'START_TRAINING': False,
  'TRAINING': {},
  'AUTO_DEPLOY': {},

  "ALLOW_EMPTY_INPUTS": True,

  'PLUGIN_LOOP_RESOLUTION': 1 / 5,  # once at 5s

  'VALIDATION_RULES': {
    **BasePlugin.CONFIG['VALIDATION_RULES'],
  },
}


class Ai4eEndToEndTrainingPlugin(BasePlugin):
  def on_init(self):
    super(Ai4eEndToEndTrainingPlugin, self).on_init()
    self.__started_gather = False
    self.__started_training = False
    self.maybe_load_previous_state()
    return

  def maybe_load_previous_state(self):
    obj = self.persistence_serialization_load()
    if obj is not None:
      self.__started_gather = obj.get('started_gather', self.__started_gather)
      self.__started_training = obj.get('started_training', self.__started_training)
    # endif obj is not None
    return

  def save_current_state(self):
    obj = {
      'started_gather': self.__started_gather,
      'started_training': self.__started_training,
    }
    self.persistence_serialization_save(obj)
    return

  def get_auto_deploy(self):
    auto_deploy = self.cfg_auto_deploy
    if auto_deploy is None:
      return None
    if 'BOX_ID' not in auto_deploy or "NODE_ADDRESS" not in auto_deploy:
      auto_deploy['BOX_ID'] = auto_deploy.get('BOX_ID', self.ee_id)
      auto_deploy['NODE_ADDRESS'] = auto_deploy.get('NODE_ADDRESS', self.node_addr)
    if 'STREAMS' not in auto_deploy:
      auto_deploy['STREAMS'] = []
    return auto_deploy

  def get_data(self):
    data_cfg = {
      **DEFAULT_DATA_CONFIG,
      **self.cfg_data
    }
    return data_cfg

  def get_training(self):
    training_cfg = {
      **DEFAULT_TRAIN_CONFIG,
      **self.cfg_training
    }
    return training_cfg

  def classes_dict(self):
    """
    The `CLASSES` parameter should be a dictionary with the following structure:
    {
      "class_1": "Class 1 description",
      "class_2": "Class 2 description",
      ...
    }
    Returns
    -------
    dict : The classes dictionary
    """
    return self.cfg_classes

  def classes_list(self):
    """
    Utility function to return the classes as a list
    Returns
    -------
    list : The classes list
    """
    return list(self.cfg_classes.keys())

  """DATA SECTION"""
  if True:
    @property
    def data_sources(self):
      return self.get_data().get('SOURCES', [])

    @property
    def data_crop_plugin_params(self):
      return self.get_data().get('CROP_PLUGIN_PARAMS', {})

    @property
    def data_cloud_path(self):
      return self.get_data().get('CLOUD_PATH', '')

    @property
    def data_train_size(self):
      return self.get_data().get('TRAIN_SIZE', 0.8)
  """END DATA SECTION"""

  """TRAINING SECTION"""
  if True:
    @property
    def training_box_id(self):
      return self.get_training().get('BOX_ID')

    @property
    def training_model_architecture(self):
      return self.get_training().get('MODEL_ARCHITECTURE')

    @property
    def training_device_load_data(self):
      return self.get_training().get('DEVICE_LOAD_DATA')

    @property
    def device_training(self):
      return self.get_training().get('DEVICE_TRAINING')

    @property
    def training_pipeline_signature(self):
      return self.get_training().get('TRAINING_PIPELINE_SIGNATURE')

    @property
    def training_grid_search(self):
      return self.get_training().get('GRID_SEARCH')

    @property
    def training_batch_size(self):
      return self.get_training().get('BATCH_SIZE')

    @property
    def training_epochs(self):
      return self.get_training().get('EPOCHS')
  """END TRAINING SECTION"""

  @property
  def dataset_object_name(self):
    return self.os_path.join(self.data_cloud_path, self.cfg_objective_name)

  """CONFIG PIPELINE SECTION"""
  if True:
    def process_data_gather_config(self, config):
      """
      Fill in the gaps in the data gather config
      """
      return {
        **config,
        'CAP_RESOLUTION': config.get('CAP_RESOLUTION', 25),
      }

    def _configured_metastream_collect_data(self):
      object_type = self.cfg_general_detector_object_type
      if not isinstance(object_type, list):
        object_type = [object_type]

      cfg_instance = {
        "INSTANCE_ID": self.get_instance_id(),
        "OBJECTIVE_NAME": self.cfg_objective_name,
        "DESCRIPTION": self.cfg_description,
        "CLOUD_PATH": self.data_cloud_path,
        "OBJECT_TYPE": object_type,
        'CLASSES': self.classes_dict(),
        'TRAIN_SIZE': self.data_train_size,
        'REWARDS': self.cfg_rewards,
        'DATASET': self.cfg_dataset,
        'CREATION_DATE': self.cfg_creation_date,
        'DATA_SOURCES': self.data_sources,
        **self.data_crop_plugin_params,
      }

      current_ai_engine = self.cfg_ai_engine
      if current_ai_engine is not None and len(current_ai_engine) > 0 and cfg_instance.get('AI_ENGINE', None) is None:
        cfg_instance['AI_ENGINE'] = current_ai_engine

      config = {
        "NAME": f"collect_data_{self.get_instance_id()}",
        "TYPE": "MetaStream",
        "COLLECTED_STREAMS": [x['NAME'] for x in self.data_sources],
        "PLUGINS": [
          {
            "SIGNATURE": "ai4e_crop_data",
            "INSTANCES": [cfg_instance]
          }
        ]
      }
      return config

    def _configured_training_pipeline(self):
      config = {
        "NAME": "training_{}".format(self.get_instance_id()),
        "TYPE": "minio_dataset",
        "STREAM_CONFIG_METADATA": {
          "DATASET_OBJECT_NAME": self.dataset_object_name,
        },
        "PLUGINS": [
          {
            "SIGNATURE": "second_stage_training_process",
            "INSTANCES": [
              {
                "INSTANCE_ID": self.get_instance_id(),
                'AI_ENGINE': f'th_training?{self.get_instance_id()}',
                'STARTUP_AI_ENGINE_PARAMS': {
                  'PIPELINE_SIGNATURE': self.training_pipeline_signature,
                  'PIPELINE_CONFIG': {
                    'CLASSES': self.classes_list(),
                    'MODEL_ARCHITECTURE': self.training_model_architecture,
                    'MODEL_NAME': self.get_instance_id(),
                    'PRELOAD_DATA': self.training_device_load_data is not None,
                    'DEVICE_LOAD_DATA': self.training_device_load_data,
                    'DEVICE_TRAINING': self.device_training,
                    'GRID_SEARCH': self.training_grid_search,
                    'BATCH_SIZE': self.training_batch_size,
                    'EPOCHS': self.training_epochs,
                    'FIRST_STAGE_TARGET_CLASS': self.cfg_general_detector_object_type,
                  },
                },
                'AUTO_DEPLOY': self.get_auto_deploy(),
                "DESCRIPTION": self.cfg_description,
                "OBJECTIVE_NAME": self.cfg_objective_name,
                "REWARDS": self.cfg_training_rewards
              }
            ]
          }
        ]
      }
      return config
  """END CONFIG PIPELINE SECTION"""

  def _process(self):
    if not self.__started_gather:
      self.P(f"Starting end to end training job {self.cfg_objective_name} with id {self.get_instance_id()}")
      for config in self.data_sources:
        processed_config = self.process_data_gather_config(config)
        self.P(f"Starting pipeline {config['NAME']} for data acquisition.")
        self.cmdapi_start_stream_by_config_on_current_box(config_stream=processed_config)
      # endfor data sources
      self.P(f"Starting metastream for data collection.")
      self.cmdapi_start_metastream_by_config_on_current_box(
        config_metastream=self._configured_metastream_collect_data()
      )
      self.__started_gather = True
      self.save_current_state()
    # endif not started gather
    if not self.__started_training and self.cfg_start_training:
      training_box_addr = self.net_mon.network_node_addr(self.training_box_id)
      self.P(f"Starting training pipeline for {self.cfg_objective_name} on {training_box_addr}")
      self._cmdapi_start_stream_by_config(
        config_stream=self._configured_training_pipeline(),
        node_address=training_box_addr
      )
      self.__started_training = True
      self.save_current_state()
    # endif not started training
    return



AI4E_MIXIN_CONFIG = {
  'REPORT_PERIOD': 10*60,
  'OBJECTIVE_NAME': None,
  'OBJECT_TYPE': ['person'],
  'CLASSES': None,

  "DESCRIPTION": "",
  "REWARDS": {},
  "DATASET": {},
  "CREATION_DATE": None,
  "DATA_SOURCES": [],

  "VALIDATION_RULES": {
  },
}


class _Ai4eMixin(object):
  def get_status_payload_additional_kwargs(self, **kwargs):
    return {}

  def get_dataset_details(self):
    return {**self.cfg_dataset}

  def get_ds_classes(self):
    classes = self.cfg_classes
    if classes is None:
      classes = self.cfg_object_type
    if not isinstance(classes, (list, dict)):
      classes = [classes]
    if not isinstance(classes, dict):
      classes = {k: k for k in classes}
      return classes
    return classes

  def get_job_status(self):
    return 'Not implemented'

  def get_status_payload_kwargs(self, **kwargs):
    res = {
      'is_status': True,
      "objective_name": self.cfg_objective_name,
      "rewards": self.cfg_rewards,
      "dataset": self.get_dataset_details(),
      "creation_date": self.cfg_creation_date,
      "data_sources": self.cfg_data_sources,
      "target": self.cfg_object_type,
      "classes": self.get_ds_classes(),
      "description": self.cfg_description,
      "job_status": self.get_job_status(),
      **self.get_status_payload_additional_kwargs(**kwargs),
    }

    return res

  def maybe_report_status(self, force=False, **kwargs):
    """
    Periodically report the status to the AI4E plugins.
    """
    if force or self.last_payload_time is None or self.time() - self.last_payload_time > self.cfg_report_period:
      status_payload_kwargs = self.get_status_payload_kwargs()
      self.add_payload_by_fields(**status_payload_kwargs)
    # endif report time
    return




import smtplib, ssl
import traceback

from naeural_core import Logger

def send_email(log, server, port, user, password, destination, msg, subject=None, sender=None, return_logs=False):
  if subject is None:
    subject = 'DecentrAI Automated message'
  if sender is None:
    sender = user
  full_message = "From: {}\nSubject: {}\n\n{}""".format(sender, subject, msg) 
    
  result = None
  svr = None
  logs = []
  try:
    if port == 25:
      # standard simple SMTP
      svr = smtplib.SMTP(server)
    else:
      context = ssl.create_default_context()
      if port == 465:
        # standard SSL
        svr = smtplib.SMTP_SSL(server, port, context=context)
      elif port == 587:
        # more secure TSL
        svr = smtplib.SMTP(server,port)
        svr.starttls(context=context) # Secure the connection
      else:
        logs.append({
          'msg' : "send_mail to '{}' failed. Unknown port {} for server '{}'".format(destination, port, server),
          'color' : 'r'
        })
  except Exception as e:
    logs.append({
      'msg' : "Exception '{}' in send_email connection:\n{}".format(
        e,
        traceback.format_exc()),
      'color' : 'r'
    })
    svr = None

  if svr is not None:
    try:
      svr.login(user, password)  
      result = svr.sendmail(
        from_addr=sender, 
        to_addrs=destination, 
        msg=full_message)  
      svr.quit()
    except Exception as e:
      logs.append({
        'msg' : "Exception '{}' in send_email delivery:\n{}".format(
          e,
          traceback.format_exc()),
        'color' : 'r'
      })

  if not return_logs:
    for dl in logs:
      log.P(dl['msg'], color=dl['color'])
    return result
  else:
    return result, logs







import tensorflow as tf
import numpy as np
import flask
import datetime

from functools import wraps
from time import sleep, time

try:
  import jwt
except ModuleNotFoundError:
  pass

from ..logger_mixins.serialization_json_mixin import NPJson


__VER__ = '0.9.2.5'

class NonReentrantModelServer:
  def __init__(self, model, model_fun, session, graph, log, fun_input, fun_output):
    self.is_tf1 = log.TF_VER[0] == '1' 
    self.graph = graph
    self.model = model
    self.model_fun = model_fun
    if not hasattr(self.model, self.model_fun):
      raise ValueError("The class {} does not have function '{}'".format(
          self.model.__class__.__name__, model_fun))
    self.session = session
    self.in_use = False
    self.fun_input_convert = fun_input
    self.fun_output_convert = fun_output
    self.version = __VER__
    self.log = log
    self.n_servings = 0
    
    
  def predict(self, inputs, client_id=None, wait_time=0, verbose=True):
    STR_MAX = 1024
    results = None
    if self.in_use:
      return
    self.in_use = True
    _ver = None
    if hasattr(self.model,'version'):
      _ver = self.model.version
    
    if verbose >= 2:
      self.log.P("")
      self.log.P("Serving prediction to client '{}' using tf {} - ipython: {}".format(
        client_id, tf.__version__, self.log.is_running_from_ipython))
      self.log.P("  Model:   {} ({}) ver:{}".format(hex(id(self.model)), self.model.__class__.__name__, _ver))
      self.log.P("  Graph:   {}".format(self.graph)) #hex(id(self.graph))))
      self.log.P("  Session: {}".format(self.session)) #hex(id(self.session)))
    #endif
    
    _error_vals = [None,None,None]
    try:
      x_inputs = self.fun_input_convert(inputs)
      ########
      if inputs is not None:
        l1 = len(inputs)
        l2 = len(x_inputs)
      else:
        l1 = None
        l2 = None
        
      __info = "input {}:{} and outputs {}:{}".format(
          type(inputs), l1, type(x_inputs), l2)
      if verbose >= 2:
        self.log.iP("Executed fun_input_convert w info: " +__info )
      ########
      
    except:
      _err_type, _err_file, _err_func, _err_line = self.log.get_error_info()
      self.log.P("ERROR {} RAISED IN fun_input_convert:".format(
          _err_type))
      _error_vals[0] = "{}".format(_err_file)
      _error_vals[1] = "{}".format(_err_type)
      _error_vals[2] = "{}".format(_err_line)
      self.log.P("ERROR INFO: {}".format(_error_vals))
      y_results = None
      
    if x_inputs is None or _error_vals[0]:
      # some clean-up
      pass
    else:
      if self.is_tf1:
        tf.keras.backend.set_session(self.session)
      
      predict_func = getattr(self.model, self.model_fun)
      try:
        if self.is_tf1:
          with self.graph.as_default():
            y_results = predict_func(x_inputs)
        else:
          y_results = predict_func(x_inputs)
        ########
        if verbose >= 2:
          self.log.iP("Executed predict_func w Inputs {} and Outputs {}".format(len(y_results), len(x_inputs)))
        ########
      except:
        _err_type, _err_file, _err_func, _err_line = self.log.get_error_info()
        self.log.P("ERROR '{}' RAISED IN PREDICTION ENGINE:".format(
            _err_type))  
        _error_vals[0] = "{}".format(_err_file)
        _error_vals[1] = "{}".format(_err_type)
        _error_vals[2] = "{}".format(_err_line)
        self.log.P("ERROR INFO: {}".format(_error_vals))
        y_results = None
                
      try:
        results = self.fun_output_convert(y_results)
        ########
        if verbose >= 2:
          self.log.iP("Executed fun_output_convert w Inputs {} and Outputs {}".format(len(y_results), len(results)))
        ########
      except:
        _err_type, _err_file, _err_func, _err_line = self.log.get_error_info()
        self.log.P("ERROR {} RAISED IN fun_output_convert:".format(
            _err_type))
        _error_vals[0] = "{}".format(_err_file)
        _error_vals[1] = "{}".format(_err_type)
        _error_vals[2] = "{}".format(_err_line)
        self.log.P("ERROR INFO: {}".format(_error_vals))
        y_results = None
        
    if results is None:
      results = {}

    if type(results) is dict:
      results['client'] = client_id  # add the client id to the json response
      
      if np.any(np.array(_error_vals) != None):
        for i,_err in enumerate(_error_vals):
          results['ERROR{}_DEBUG'.format(i)] = _err
        results['INFO_DEBUG'] = "Please send this data to dev team. Thank you."
      if _ver is not None:
        _key = "{}".format(self.model.__class__.__name__)
        results[_key] = _ver
      results['log'] = self.log.version
      if hasattr(self.log, 'timeseries_benchmarker'):
        results['ts'] = self.log.timeseries_benchmarker.version
    #endif
    
    str_res = str(results)
    str_res = str_res if len(str_res) <=STR_MAX else str_res[:STR_MAX]+'...'
    self.n_servings += 1
    if verbose >= 1:
      self.log.P("Response #{} for client:'{}': {}".format(self.n_servings, client_id, str_res))
      self.log.P(" ")
    c_time = time()
    while (c_time + wait_time) > time():
      sleep(1)
    self.in_use = False
    return results

###############################################################################

class SimpleFlaskModelServer(object):
  """
  Simple engine that allows the quick'n'dirty operationalization of a model. Ok,
  maybe not that dirty ;)
  
  Initialization:

    `model` : tf/keras trained/loaded model or a class that has a `.predict` function
    
    `fun_input`  : callback that receives a specific input json and transforms into a
                 model input. Ideally the input json should also contain a `'client'` 
                 key and value
                
    `fun_output` : callback that receives model output and prepares output json
    
    `log`  : mandatory Logger object
    
    `host` : host default to local
    
    `port` : port default 5000
  
  """
  app = None
  def __init__(self, model, log, endpoint_name='/analyze',
               fun_input=None, fun_output=None, 
               host='127.0.0.1', port=5000, predict_function='predict',
               db_file=None, signup=True, verbose=True, workers=1):
    self.__version__ = __VER__
    self.__name__ = 'MSRV'
    self.host = host
    self.port = port
    self.log = log
    self.model = model
    self.predict_function_name = predict_function
    self.authentication = db_file is not None
    self.model_servers = None
    self.verbose = verbose
    self.nr_workers = workers
    assert self.nr_workers > 0, "There should be minimum 1 worker."
    
    if fun_input is None:
      fun_input = lambda x:x
      self.log.P("Using identity input [fun_input]")
    if fun_output is None:
      fun_output = lambda x:x
      self.log.P("Using identity output [fun_output]")

    self.fun_input = fun_input
    self.fun_output = fun_output
    self.app = flask.Flask('SimpleFlaskModelServer')
    self.app.json_encoder = NPJson
    self.create_analyze_endpoint(endpoint_name)

    str_auth_log = '(without auth)'
    if self.authentication:
      self.app.config['SECRET_KEY'] = 'thisissecretkey'
      self.app.config['SQLALCHEMY_DATABASE_URI'] = 'sqlite:////{}'.format(db_file)
      self.app.config['SQLALCHEMY_TRACK_MODIFICATIONS'] = True

      from flask_sqlalchemy import SQLAlchemy
      self.db = SQLAlchemy(self.app)
      db = self.db

      class Users(db.Model):
        id = db.Column(db.Integer, primary_key=True)
        public_id = db.Column(db.Integer)
        name = db.Column(db.String(50))
        password = db.Column(db.String(50))
        admin = db.Column(db.Boolean)

      self.Users = Users
      if signup:
        self.create_signup_user_endpoint()
      #endif
      self.create_login_user_endpoint()
      str_auth_log = '(with auth)'
    #endif

    _logo = "SimpleFlaskModelServer v{} {} started on '{}:{}'".format(
        self.__version__, str_auth_log, host, port)

    lead = 5
    _logo = " " * lead + _logo
    s2 = (len(_logo) + lead) * "*"
    self.log.P("")
    self.log.P(s2)
    self.log.P("")
    self.log.P(_logo)
    self.log.P("")
    self.log.P(s2)
    self.log.P("")


  def run(self):
    self.app.run(host=self.host, port=self.port)


  def create_signup_user_endpoint(self):
    import uuid
    from werkzeug.security import generate_password_hash

    def endp():
      data = flask.request.get_json()

      hashed_password = generate_password_hash(data['password'], method='sha256')

      new_user = self.Users(public_id=str(uuid.uuid4()), name=data['user'], password=hashed_password, admin=False)
      self.db.session.add(new_user)
      self.db.session.commit()
      return flask.jsonify({'message': 'registered successfully'})

    self.app.add_url_rule(rule='/signup',
                          endpoint='SignupEndpoint',
                          view_func=endp,
                          methods=['GET', 'POST'])


  def create_login_user_endpoint(self):
    from werkzeug.security import check_password_hash

    def endp():
      auth = flask.request.authorization

      if not auth or not auth.username or not auth.password:
        return flask.make_response('could not verify', 401, {'WWW.Authentication': 'Basic realm: "login required"'})

      user = self.Users.query.filter_by(name=auth.username).first()

      if check_password_hash(user.password, auth.password):
        token = jwt.encode(
          {'public_id': user.public_id, 'exp': datetime.datetime.utcnow() + datetime.timedelta(minutes=30)},
          self.app.config['SECRET_KEY'])
        return flask.jsonify({'token': token.decode('UTF-8')})

      return flask.make_response('could not verify', 401, {'WWW.Authentication': 'Basic realm: "login required"'})

    self.app.add_url_rule(rule='/login',
                          endpoint='LoginEndpoint',
                          view_func=endp,
                          methods=['GET', 'POST'])


  def create_analyze_endpoint(self, endpoint_name='/analyze'):
    if self.log.TF_VER[0] == '1':
      self.session = tf.keras.backend.get_session()
      self.graph = tf.get_default_graph()
    else:
      self.session = None
      self.graph = None
    
    self.model_servers = []
    
    for _ in range(self.nr_workers):
      model_server = NonReentrantModelServer(model=self.model, 
                                             model_fun=self.predict_function_name,
                                             session=self.session,
                                             graph=self.graph,
                                             fun_input=self.fun_input,
                                             fun_output=self.fun_output,
                                             log=self.log)
    
      self.model_servers.append(model_server)
    #endfor
    
    self.log.P("Created {} workers (NonReentrantModelServer)".format(self.nr_workers))
    
    endp = SimpleFlaskModelServer.AnalyzeEndpoint(outer=self)
    if self.authentication:
      view_func = endp.call_with_token_required
    else:
      view_func = endp.call_without_token_required
    self.app.add_url_rule(rule=endpoint_name, 
                          endpoint="ModelEndpoint", 
                          view_func=view_func,
                          methods = ['GET', 'POST','OPTIONS']
                          )

  ###############################################################################

  class AnalyzeEndpoint(object):
    def __init__(self, outer):
      self.outer = outer
      self.model_servers = self.outer.model_servers
      self.model_server_in_use = np.zeros(shape=(len(self.model_servers),),
                                          dtype=np.int32)
      self.version = self.model_servers[0].version
      self.counter = 0
      self.verbose = self.outer.verbose
      return
    
    def _lock_worker(self, wid):
      self.model_server_in_use[wid] = 1
      return
    
    def _unlock_worker(self, wid):
      self.model_server_in_use[wid] = 0
      
    def _find_unlocked_worker(self):
      unlocked_workers = np.where(self.model_server_in_use == 0)[0]
      
      if unlocked_workers.shape[0] == 0:
        return
      
      wid = int(np.random.choice(unlocked_workers))
      self._lock_worker(wid)
      
      return wid

    def wait_predict(self, data, client_id=None, wait_time=0):
      wid = self._find_unlocked_worker()
      if wid is None:      
        self.outer.log.P("  All model servers in use. '{}' waiting...".format(client_id))
        while wid is None:
          sleep(1)
          wid = self._find_unlocked_worker()
          
        self.outer.log.P("  Waiting done for '{}'. Worker {} is now free".format(client_id, wid))
      
      # now worker is locked...
      model_server = self.model_servers[wid]
      answer = model_server.predict(data,
                                    client_id=client_id,
                                    wait_time=wait_time,
                                    verbose=self.verbose)
      self._unlock_worker(wid)
      return answer, wid

    def token_required(f):
      @wraps(f)
      def decorator(self, *args, **kwargs):
        token = None

        if 'x-access-tokens' in flask.request.headers:
          token = flask.request.headers['x-access-tokens']

        if not token:
          return flask.jsonify({'message': 'a valid token is missing'})

        try:
          _ = jwt.decode(token, self.outer.app.config['SECRET_KEY'])
        except:
          return flask.jsonify({'message': 'token is invalid'})

        return f(self, *args, **kwargs)

      return decorator

    @token_required
    def call_with_token_required(self, *args):
      return self.call(*args)

    def call_without_token_required(self, *args):
      return self.call(*args)

    def call(self, *args):
      self.counter += 1
      ccall = self.counter
      request = flask.request
      method = flask.request.method
      args_data = request.args
      form_data = request.form
      json_data = request.json
      params = None
      client = 'unk'

      if method == 'GET':
        # parameters in URL
        base_params = args_data
      else:
        # parameters in form
        base_params = form_data
        if len(base_params) == 0:
          # params in json?
          base_params = json_data

      if base_params is not None:
        if 'client' in base_params.keys():
          client = base_params['client']
        params = dict(base_params)

      
      if self.verbose >= 1:
        self.outer.log.P("Received '{}' request {} from client '{}' params: {}".format(
          method, ccall, client, params))

      ######
      # print("args_data: ", args_data)
      # print("form_data: ", form_data)
      # print("json_data: ", json_data)
      ######
      
      wid = -1
      if method != 'OPTIONS':
        answer, wid = self.wait_predict(params, client_id=client, wait_time=0)
      else:
        self.outer.log.P("Received OPTIONS request")
        answer = {}
      if answer is None:
        jresponse = flask.jsonify({
          "ERROR": "input json does not contain right info or other error has occured",
          "client": client,
          "call_id": ccall,
          "input": base_params,
          "server": self.version})
      else:
        if type(answer) is dict:
          answer['call_id'] = ccall
          answer['server'] = self.version
          answer['wid'] = wid
          jresponse = flask.jsonify(answer)
        elif type(answer) is str:
          jresponse = flask.make_response(answer)

      jresponse.headers["Access-Control-Allow-Origin"] = "*"
      jresponse.headers["Access-Control-Allow-Methods"] = "POST, GET, OPTIONS, DELETE"
      jresponse.headers["Access-Control-Allow-Headers"] = "Content-Type"
      return jresponse

  ###############################################################################


if __name__ == '__main__':
  
  class FakeLogger:
    def __init__(self, **kwargs):
      return
    
    def P(self, s):
      print("LOG: " + s, flush=True)
  
  def inp_proc(data):
    if 'input_value' not in data.keys():
      print("ERROR: input json does not contain data")
      return None
    # get string
    s = data['input_value']     
    # create dict
    d = [chr(x) for x in range(5000)] 
    # tokenize
    t = [d.index(x) for x in s] 
    # batchfy
    np_t = np.array(t).reshape((1,-1))
    return np_t
  
  def out_proc(data):
    # create dict
    d = [chr(x) for x in range(5000)]
    # select first obs
    t = data[0]
    t = t.ravel().astype(int)
    # get string 
    c = [d[x] for x in t]    
    s = "".join(c)
    return {'output_value':str(s)}
  
  l = FakeLogger(lib_name='MSRVT', no_folders_no_save=True)
    
  # model gets a sequnce
  tf_inp = tf.keras.layers.Input((None,))
  # adds a value to each token in all sequences
  tf_x = tf.keras.layers.Lambda(lambda x: x+1)(tf_inp)
  # returns modified input
  m = tf.keras.models.Model(tf_inp, tf_x)
  


  
  class FakeModel(object):
    def __init__(self, logger, fn_model):
      self.logger = logger
      self.fn_model = fn_model
      self.version = '0.0.1'
    
    def predict(self, usr_input):
      self.logger.P('Predicting on usr_input: {}'.format(usr_input))
      
      res = '{}+1={} PREDICTED'.format(usr_input,int(usr_input)+1)
      
      return res
    
  dummy_model = FakeModel(l, 'path/to/model.h5')
    
  def dummy_inp_proc(data):
    if 'input_value' not in data.keys():
      print("ERROR: input json does not contain data")
      return None
  
    s = data['input_value']      

    return s
  
  def dummy_out_proc(data):
    return {'output_value':str(data)}
  
  dummy_server = SimpleFlaskModelServer(model=dummy_model, 
                               fun_input=dummy_inp_proc, 
                               fun_output=dummy_out_proc,
                               log=l,
                               port=5000)
  
  dummy_server.run()


from ratio1 import BaseCodeChecker

class _CodeExecutorMixin(BaseCodeChecker):
  """
  IMPORTANT: this class must be a mixin 
  """
  def __init__(self):
    super(_CodeExecutorMixin, self).__init__()
    return
  


import gc
import abc
import random
import multiprocessing
import inspect

import torch as th
import torchvision.transforms as T
import numpy as np
from typing import Union, List, Tuple, Any
from naeural_core import DecentrAIObject
from naeural_core import Logger
from naeural_core.local_libraries.nn.th.training_utils import get_shallow_dataset, to_device

class IterableDataset(th.utils.data.IterableDataset):

  def __init__(self, observations : List,
               labels: List,
               is_train_dataset : bool,
               are_observations_loaded : bool,
               load_x_and_y_callback,
               **kwargs):
    super(IterableDataset, self).__init__(**kwargs)
    self._observations = observations
    self._labels = labels
    self._is_train_dataset = is_train_dataset
    self._are_observations_loaded = are_observations_loaded
    self._indexes = list(range(len(self._observations)))

    self.load_x_and_y_callback = load_x_and_y_callback
    self.startup()
    return

  def startup(self):
    return

  def _shuffle(self):
    random.shuffle(self._indexes)
    return

  def _get_worker_indexes(self):
    ## TODO: Docstrings
    worker = th.utils.data.get_worker_info()
    worker_id = worker.id if worker is not None else -1

    if worker_id != -1:
      split_size = len(self._indexes) // worker.num_workers
      worker_indexes = range(worker_id * split_size, (worker_id + 1) * split_size)
    else:
      worker_indexes = range(len(self._indexes))
    #endif
    return worker_indexes

  def __len__(self):
    return len(self._indexes)

  def update_dataset(self, **kwargs):
    return

  def __iter__(self):
    self._shuffle()
    worker_indexes = self._get_worker_indexes()

    for i in worker_indexes:
      idx = self._indexes[i]

      if not self._are_observations_loaded:
        x,y = self.load_x_and_y_callback(self._observations, self._labels, idx)
      else:
        x = self._observations[idx]
        y = self._labels[idx]
      #endif

      ret = [x, y, idx]
      # if not self._is_train_dataset:
      #   ret.append(idx)

      yield tuple(ret)
    #endfor

  def release_dataset(self):
    del self._observations
    del self._labels
    gc.collect()
    th.cuda.empty_cache()
    return

class BaseDataLoaderFactory(DecentrAIObject, abc.ABC):

  def __init__(self, log : Logger,
               batch_size : int,
               path_to_dataset : str, data_subset_name : str,
               load_observations: bool = True,
               load_device : str = 'cpu',
               training_device : str = 'cuda:0',
               files_extensions : Union[List[str], str] = None,
               num_workers : int = None,
               **kwargs):
    """
    Data loader factory. Should be initialized one for each subset (train, dev, test).

    Parameters:
    -----------
    batch_size: int
      Batch size for training time

    path_to_dataset: str
      Local path to the subset of the data.

    dataset_subset_name : str
      The name of the subset of the data.
      Could be "train", "dev" or "test"

    load_observations: bool, optional
      Flag that determines if the observations will be fully loaded to be available in-memory at training time.
      The functionality of pre-loading the data before training speeds up the training process. If the dataset is
      too big to be accomodated on gpu, it is ok to pre-load it in the RAM memory. Pre-loading it directly on the
      gpu that is responsible with the training brings the best speedup.
      The default value is True.

    load_device: str, optional
      The device where the data will be loaded if `load_observations` is True.
      For gpu devices use "cuda:<index>".
      The default value is "cpu:0"

    training_device: str, optional
      The device where the training will happen.
      For gpu devices use "cuda:<index>".
      The default value is "cpu:0"

    files_extentions: Union[List[str], str], optional
      Which files from the dataset to be considered (based on their extensions).
      For example, a dataset may have images (.jpg, .png) and metadata files (.json / .xml) but for a particular model
      (let's say autoencoding) only the images are needed. In this case, set `file_extensions` to be `['.jpg', '.png']`
      The default value is None (i.e. will take into consideration all the files)
    """
    self._load_observations = load_observations
    self._batch_size = batch_size
    self._path_to_dataset = path_to_dataset
    self._data_subset_name = data_subset_name.lower()
    assert self._data_subset_name in ['train', 'dev', 'test']
    self._is_train_dataset = (self._data_subset_name == 'train')
    self.load_device = load_device
    self.training_device = training_device

    self._num_workers = 0 if self._load_observations else (
      num_workers if num_workers is not None else min(int(0.7 * multiprocessing.cpu_count()), 8)
    )
    self._pin_memory = self._num_workers != 0
    self._files_extensions = files_extensions

    self.dataset = None
    self.data_loader = None
    self.dataset_info = None

    self.lst_on_load_preprocess_definitions = []
    self.lst_before_forward_preprocess_definitions = []
    self._on_load_preprocess = None
    self._before_forward_preprocess = None
    self._augmentation = None
    self._collate_callback = getattr(self, "_collate_fn", None)

    self.data_loader_kwargs = {}
    super(BaseDataLoaderFactory, self).__init__(log=log, **kwargs)
    return

  def startup(self):
    super().startup()
    self.__name__ = self.log.name_abbreviation(self.__class__.__name__) + '_{}'.format(self._data_subset_name)

    self.dataset_info = get_shallow_dataset(
      log=self.log,
      path_to_dataset=self._path_to_dataset,
      extensions=self._files_extensions
    )

    self.dataset_info['is_train_dataset'] = self._is_train_dataset

    self.lst_on_load_preprocess_definitions = self._lst_on_load_preprocess()
    self.lst_before_forward_preprocess_definitions = self._lst_right_before_forward_preprocess()
    lst_on_load_preprocess = [x[0](**x[1]) for x in self.lst_on_load_preprocess_definitions]
    lst_before_forward_preprocess = [x[0](**x[1]) for x in self.lst_before_forward_preprocess_definitions]

    self._on_load_preprocess = T.Compose(lst_on_load_preprocess)
    self._before_forward_preprocess = T.Compose(lst_before_forward_preprocess)
    self.P("{} data factory initialized".format(self._data_subset_name.upper()), color='g')
    return

  @property
  def preprocess_definitions(self):
    return self.lst_on_load_preprocess_definitions + self.lst_before_forward_preprocess_definitions

  def P(self, s, **kwargs):
    super().P(s, prefix=True, **kwargs)
    return

  def _get_iterable_dataset_params(self) -> dict:
    observations_not_loaded, labels_not_loaded = self._get_not_loaded_observations_and_labels()
    assert len(observations_not_loaded) == len(labels_not_loaded)
    nr_observations = len(observations_not_loaded)

    observations, labels = [], []
    if self._load_observations:
      self.P("Loading the whole dataset ({} observations) to device={} ...".format(len(observations_not_loaded), self.load_device))
      for idx in range(nr_observations):
        x,y = self.load_x_and_y_callback(observations_not_loaded, labels_not_loaded, idx)
        observations.append(x)
        labels.append(y)

        if self.log.is_main_thread:
          print(
            '\r  Load progress: {}/{} ({:.2f}%)'.format(idx + 1, nr_observations, 100 * (idx + 1) / nr_observations),
            flush=True, end=''
          )
        #endif
      #endfor
      self.P("  Dataset loaded.")
      self.log.show_timers(selected_sections=['{}_preload'.format(self.__name__)])
    else:
      observations = observations_not_loaded
      labels = labels_not_loaded
    #endif

    dct = {
      'class_ref' : IterableDataset,
      'observations' : observations,
      'labels' : labels,
      'other_params':{
        **self.data_loader_kwargs
      }
    }
    return dct

  @abc.abstractmethod
  def _lst_on_load_preprocess(self) -> List[Tuple[Any, dict]]:
    """
    List of preprocessing techniques that will be performed at loading time.

    For example, if the images in the dataset do not have the same H,W, one may want to resize all the images to
    specific H,W.

    All the preprocessing techniques should be wrapped in classes that expose a __call__ method.
    (please see libraries.nn.th.image_dataset_stage_preprocessor).

    All these preprocess methods are automatically packed into a th transformation (T.Compose)

    Returns:
    -------
    List[Tuple[Any, dict]]
      for each preprocessing technique, a tuple with: a) the class reference and b) the class __init__ kwargs
      should be provided
    """
    raise NotImplementedError

  @abc.abstractmethod
  def _lst_right_before_forward_preprocess(self) -> List[Tuple[Any, dict]]:
    """
    List of preprocessing techniques that will be performed at right before model forward.

    For example, the images min-max normalization.

    The reason why this functionality is separated from the functionality exposed in `_lst_on_load_preprocess`
    is that between resize and min-max normalization, one may want to define some augmentations for training time.
    (please see libraries.nn.th.training.callbacks.base, method `_lst_augmentations`)

    All the preprocessing techniques should be wrapped in classes that expose a __call__ method.
    (please see libraries.nn.th.image_dataset_stage_preprocessor)

    All these preprocess methods are automatically packed into a th transformation (T.Compose)

    WARNINIG: None of these preprocess should be augmentations such as ColorJitter, because this is also applied during eval phase

    Returns:
    -------
    List[Tuple[Any, dict]]
      for each preprocessing technique, a tuple with: a) the class reference and b) the class __init__ kwargs
      should be provided
    """
    raise NotImplementedError

  @abc.abstractmethod
  def _get_not_loaded_observations_and_labels(self) -> Tuple[Union[List, np.ndarray], Union[List, np.ndarray]]:
    """
    Defines the sources of the observations and labels (before loading them in memory). Most of the time the
    observations are paths to some data and the labels are the actual labels as they are not very memory-consuming.
    The output of this method is passed to `_load_x_and_y` for actually loading in memory an observation and a label.

    Returns:
    -------
    (observations, labels): Tuple[Union[List, np.ndarray], Union[List, np.ndarray]]
      The sources of the observations and labels
    """
    raise NotImplementedError

  @abc.abstractmethod
  def _load_x_and_y(self, observations, labels, idx) -> Tuple[Union[List, np.ndarray], Union[List, np.ndarray]]:
    """
    Provides an actual in-memory loading logic for a certain observation and label.
    Do not bother with th.tensors and device transfers, because all these things are performed automatically.

    Parameters:
    ----------
    observations: Union[List, np.ndarray]
      The first element in the tuple returned by `_get_not_loaded_observations_and_labels`

    labels: Union[List, np.ndarray]
      The second element in the tuple returned by `_get_not_loaded_observations_and_labels`

    idx: int
      The observation and label index that should be loaded

    Returns:
    -------
    (x, y): Tuple[Union[List, np.ndarray], Union[List, np.ndarray]]
      np.ndarray tensors for a certain input and a corresponding output
      For multi-input and multi-output, there will be a list of np.ndarray tensors
    """
    raise NotImplementedError

  @abc.abstractmethod
  def _preprocess(self, x, y, transform):
    """
    Defines the actual preprocessing step for some input and its output
    Most of the time the output should not be preprocessed, so it just has to be returned without any modification.

    Parameters:
    ----------
    x: Union[List, th.tensor]
      The input

    y: Union[List, th.tensor]
      The output

    transform: T.Compose
      A transformation that should be applied. This is the compose generated from the list of preprocess
        methods, either for load or forward

    Returns:
    -------
    (x, y): Tuple[Union[List, np.ndarray], Union[List, np.ndarray]]
      Transformed x and y
    """
    raise NotImplementedError

  def update_data_factory(self, **kwargs):
    if inspect.getsource(self.dataset.update_dataset) == inspect.getsource(IterableDataset.update_dataset):
      raise NotImplementedError("Please implement method `update_dataset` in `IterableDataset` child")

    self.dataset.update_dataset(**kwargs)
    self.data_loader = th.utils.data.DataLoader(
      self.dataset,
      batch_size=self._batch_size,
      num_workers=self._num_workers,
      pin_memory=self._pin_memory
    )

  def load_x_and_y_callback(self, observations, labels, idx) -> Tuple[Union[List, th.tensor], Union[List, th.tensor]]:
    self.log.start_timer('one_image', section='{}_preload'.format(self.__name__))

    self.log.start_timer('load_x_and_y', section='{}_preload'.format(self.__name__))
    x, y = self._load_x_and_y(observations, labels, idx)
    self.log.end_timer('load_x_and_y', section='{}_preload'.format(self.__name__))

    self.log.start_timer('to_device', section='{}_preload'.format(self.__name__))
    x, y = self.__to_device(x, y)
    self.log.end_timer('to_device', section='{}_preload'.format(self.__name__))

    self.log.start_timer('preprocess_on_load', section='{}_preload'.format(self.__name__))
    x, y = self._preprocess_on_load(x, y)
    self.log.end_timer('preprocess_on_load', section='{}_preload'.format(self.__name__))

    self.log.end_timer('one_image', section='{}_preload'.format(self.__name__))
    return x,y

  @to_device('load_device')
  def __to_device(self, x, y):
    return x, y

  def _preprocess_on_load(self, x, y):
    x,y = self._preprocess(x,y,self._on_load_preprocess)
    return x,y

  @to_device('training_device')
  def preprocess_before_forward(self, x, y):
    x,y = self._preprocess(x, y, self._before_forward_preprocess)
    return x,y

  def create(self):
    self.P("Preparing the '{}' dataset ...".format(self._data_subset_name), color='g')
    # here can be added some logic for automatically determining whether the dataset will be fully loaded or workers will be employed, instead of using `load_observations`

    dct = self._get_iterable_dataset_params()
    assert 'class_ref' in dct
    assert 'observations' in dct
    assert 'labels' in dct

    self.dataset = dct['class_ref'](
      observations=dct['observations'],
      labels=dct['labels'],
      is_train_dataset=self._is_train_dataset,
      are_observations_loaded=self._load_observations,
      load_x_and_y_callback=self.load_x_and_y_callback,
      **dct.get('other_params', {}),
    )

    self.data_loader = th.utils.data.DataLoader(
      self.dataset,
      batch_size=self._batch_size,
      num_workers=self._num_workers,
      pin_memory=self._pin_memory,
      collate_fn=self._collate_callback
    )
    return

  def release_device(self):
    self.dataset.release_dataset()
    del self.data_loader
    gc.collect()
    th.cuda.empty_cache()
    return




import abc

from datetime import datetime
from naeural_core import Logger
from naeural_core import DecentrAIObject

class BaseTestingPlugin(DecentrAIObject):

  def __init__(self, log : Logger, config=None, owner=None, **kwargs):
    self.y_hat = None
    self.exceptions = []
    self.config = config or {}
    self.owner = owner
    super(BaseTestingPlugin, self).__init__(log=log, **kwargs)
    return

  @abc.abstractmethod
  def _register_payload(self, payload):
    raise NotImplementedError()

  def register_payload(self, payload):
    try:
      self._register_payload(payload)
    except Exception as e:
      msg = "Could not register payload due to the following error:\n{}".format(e)
      self.exceptions.append(msg)
      self.P(msg, color='r')

    return




#global dependencies


#local dependencies

from naeural_core.business.base import CVPluginExecutor as BaseClass

__VER__ = '2.0.1'

_CONFIG = {
  **BaseClass.CONFIG,

  # if this is set to true the on-idle will be triggered continously the process
  # thus at a one-time request it might not get a real image
  'RUN_WITHOUT_IMAGE'           : False, # DO NOT CHANGE THIS! This is a special plugin that needs to run with image
  'ALLOW_EMPTY_INPUTS'          : False, # DO NOT CHANGE THIS! This is a special plugin that needs to run with image
  'MAX_INPUTS_QUEUE_SIZE'       : 1,


  "NR_WITNESSES"                : 1200,  
  "PROCESS_DELAY"               : 1,

   
  "SIMPLE_WITNESS"              : False, # must not send simple due to witness marking needed
  "DEBUG_ON_WITNESS"            : False, # no bars!
  
  "INCREASED_PROCESS_DELAY"      : 3,
  
  "VIEW_SCENE_DEBUG_LOG"        : False,
  
  
  "MAX_PROCESS_DELAY"           : 10,
  
  
  'VALIDATION_RULES': {
    **BaseClass.CONFIG['VALIDATION_RULES'],
    
    "NR_WITNESSES" : {
      "TYPE"          : "int",
      "DESCRIPTION"   : "How many video stream witnesses should be sent before stopping. Set this to a low number for receivinf just the initial witnesses",
      "MIN_VAL"       :  3,
      "MAX_VAL"       :  1e10,
    },    
    
  },

}

class ViewScene01Plugin(BaseClass):
  CONFIG = _CONFIG
  def __init__(self, **kwargs):
    self.__force_send_image = False
    self.__force_send_image_prev_delay = None
    self.__view_scene_command_data = None
    super(ViewScene01Plugin, self).__init__(**kwargs)
    return

  
  def startup(self):
    super().startup()
    self.__witness_count = 0
    self.__last_valid_image = None
    self.__last_valid_time = None
    self.__last_valid_timestamp = None
    self.__sending_previous_image = False
    return


  def _draw_witness_image(self, img, **kwargs):
    if self.__sending_previous_image:
      self.P("Preparing previous image")
      img = self._painter.alpha_text_rectangle(
        image=img, 
        top=100,
        left=1, 
        text="Archive image from {} ({:.1f}s ago)".format(self.__last_valid_time, self.time() - self.__last_valid_timestamp),  
        color=self.ct.DEEP_RED,
        size=1.5
      )
    return img      
  
  def __force_run(self):
    self.__force_send_image = True
    self.__force_send_image_prev_delay = self.cfg_process_delay
    self.config_data['PROCESS_DELAY'] = 0.5
    self.config_data['RUN_WITHOUT_IMAGE'] = True
    self.config_data['ALLOW_EMPTY_INPUTS'] = True
    return
  
  def __reset_force_run(self):
    self.__force_send_image = False
    self.config_data['PROCESS_DELAY'] = self.__force_send_image_prev_delay
    self.config_data['RUN_WITHOUT_IMAGE'] = False
    self.config_data['ALLOW_EMPTY_INPUTS'] = False
    return
  
  def on_command(self, data, **kwargs):
    """
    Although we could send a witness image on command using self.add_payload() 
    we prefer to postpone the normal process and send the image on the next process hit. 
    """

    self.__view_scene_command_data = data
    if self.__last_valid_timestamp is None:
      last_msg = "No previous image - command will probably fail!"
      color = 'r'
    else:
      last_msg = "Last img {:.1f}s ago".format(self.time() - self.__last_valid_timestamp)
      color = 'b'

    # set to run no matter what
    self.__force_run()
    # end set to run no matter what

    msg = "View scene on-command. Forcing delay {:.1f}. Current: {:.1f}. IsPostponed: {}. {}".format(
      self.cfg_process_delay,
      self.__force_send_image_prev_delay,
      self.is_process_postponed,
      last_msg,
    )
    self.P(msg, color=color )
    return


  def _process(self):
    # triggered only if there is a image!
    
    payload = None
    curr_delay = None
    new_delay = None
    
    img = self.dataapi_image()
    self.__sending_previous_image = False

    force_send_needed = self.__force_send_image
    can_send_images = self.__witness_count < self.cfg_nr_witnesses
    send_image_now = force_send_needed or can_send_images
    
    if self.cfg_nr_witnesses >= 60 and self.__witness_count < self.cfg_nr_witnesses:
      self.P("WARNING! High number of witnesses requested. Sending  {}/{}".format(
        self.__witness_count, self.cfg_nr_witnesses), color='r'
      )
    
    # Stage 1: if need be we get an old image
    if force_send_needed:
      curr_delay = self.cfg_process_delay          
      new_delay = self.__force_send_image_prev_delay
      
      # reset to default behaviour
      self.__reset_force_run()
      # end reset to default behaviour
      
      self.P("Forcing image send due to on-command {}.".format(
        "using CURRENT image" if img is not None else "using PREVIOUS last valid image"
      ))
      if img is None:
        self.P("  No image to send - trying to use last valid image")
        if self.__last_valid_image is not None:
          img = self.__last_valid_image
          self.P("  Using last valid image from: {} ({:.1f}s old)".format(
            self.__last_valid_time, self.time() - self.__last_valid_timestamp)
          )
          self.__sending_previous_image = True
        else:
          msg = "ViewScene request failed: No last valid image to send"
          self.P(msg, color='r')
          self._create_abnormal_notification(
            msg=msg,
          )
        #endif we have a last valid image or not
      #endif no image but we have a command to send an image
    #endif we have a force send command
    
    
    # Stage 2: check if we have a image to send
    if img is not None:
      if not self.__sending_previous_image:
        self.__last_valid_image = img        
        self.__last_valid_time = self.time_to_str()
        self.__last_valid_timestamp = self.time()      
      #endif we have a new image

      self.int_cache['runs'] += 1
      if self.int_cache['runs'] == 1:
        self.P('First image at process: {} and exec: {}'.format(
          self.current_process_iteration, 
          self.current_exec_iteration)
        )
      elif self.cfg_view_scene_debug_log:
        self.P("Witness {} at proc: {} exec: {}, out.wrk.shft: {}, laps: {}".format(
          self.int_cache['runs'],
          self.current_process_iteration, 
          self.current_exec_iteration,
          self.outside_working_hours,
          self.loop_timings[-5:]
          )
        )
      #endif debug run

      if send_image_now:
        additional_data = None
        if force_send_needed:
          self.P("Forced image send. Delay {:.1f} -> {:.1f}".format(
            curr_delay, new_delay), color='r'
          )
          additional_data = self.__view_scene_command_data
          self.__view_scene_command_data = None
        #endif we need to restore proc delay due to recent incoming command
        
        self.__witness_count += 1
        if self.cfg_increased_process_delay > self.cfg_process_delay:
          self.config_data['PROCESS_DELAY'] = self.cfg_increased_process_delay
        #endif maybe we need to increase the process delay
                  
        np_img = self.get_witness_image(img=img)    
        dct_payload = dict(
          status="Base video camera witness",
          img=np_img,
          img_witness_id=self.__witness_count,
          img_remaining_witnesses=self.cfg_nr_witnesses - self.__witness_count,          
        )
        if additional_data is not None:
          dct_payload['COMMAND_PARAMS'] = additional_data
        #endif we have additional data to send
          
        payload = self._create_payload(
          **dct_payload
        )        
      elif self.cfg_close_pipeline_when_done:
        self.cmdapi_archive_pipeline()     
      #endif normal send & forced send
    #endif we have an image to send
    return payload



import pandas as pd
import json

class _DataframeLoaderMixin(object):
  """
  Mixin for loading a csv (decoding it from its json representation or reading it).
  This mixin is used by `plugins.data.a_dummy_dataframe` and `plugins.data.a_dummy_dataframe_map_reduce`
  """
  def __init__(self):
    super(_DataframeLoaderMixin, self).__init__()
    return

  def dataframe_load(self, source=None):
    # receives a dataframe (json representation or path for read_csv) as part 
    # of "STREAM_CONFIG_METADATA" or direclty as "URL"
    if source is None:
      source = self.cfg_stream_config_metadata.get('DF_SOURCE')
    if source is None:
      source = self.cfg_url

    try:
      json_repr = json.loads(source)
    except json.decoder.JSONDecodeError:
      json_repr = None

    if json_repr is not None:
      return pd.DataFrame(json_repr)
    else:
      return pd.read_csv(source)


# global dependencies
import os
from time import time

# local dependencies

from naeural_core import Logger

CONFIG_MINIO = {
    "ENDPOINT"   : 'MINIO_ENDPOINT',
    "ACCESS_KEY" : 'MINIO_ACCESS_KEY',
    "SECRET_KEY" : 'MINIO_SECRET_KEY',
    "BUCKET_NAME": "MINIO_BUCKET_NAME"
}

if __name__ == '__main__':
  log = Logger(
    lib_name='SBTST',
    base_folder='.',
    app_folder='_local_cache',
    config_file='config_startup.txt',
    max_lines=1000,
    TF_KERAS=False
  )

  path_fld = r'C:\Users\ETA\Downloads\blur_jobs_new\blur_jobs\9b14f110-6716-448f-a096-357f237bb7e8'

  files = [
    'original_detections.txt'
  ]

  dct_files = {}
  for name in files:
    print('Uploading {}'.format(name))
    start = time()
    url = log.minio_upload(
      file_path=os.path.join(path_fld, name),
      endpoint=CONFIG_MINIO['ENDPOINT'],
      access_key=CONFIG_MINIO['ACCESS_KEY'],
      secret_key=CONFIG_MINIO['SECRET_KEY'],
      bucket_name=CONFIG_MINIO['BUCKET_NAME'],
      object_name='/blur_tests/{}'.format(name),
      debug=True
    )
    stop = time()
    dct_files[name] = {
      'URL': url,
      'TIME': stop - start
    }
  #endfor

  for k, v in dct_files.items():
    log.p('* {}: {}'.format(k, v['URL']))




from naeural_core.business.default.web_app.fast_api_web_app import FastApiWebAppPlugin as BasePlugin

__VER__ = '0.0.0.0'

_CONFIG = {
  **BasePlugin.CONFIG,
  'PROCESS_DELAY': 5,

  'NGROK_DOMAIN': None,
  'NGROK_EDGE_LABEL': None,

  'PORT': None,
  'ASSETS': None,
  'STATIC_DIRECTORY': '.',
  'DEFAULT_ROUTE': None,

  'VALIDATION_RULES': {
    **BasePlugin.CONFIG['VALIDATION_RULES']
  },
}


class GenericHttpServerPlugin(BasePlugin):
  """
  A plugin which handles a Basic http server web app hosted through FastAPI.
  """

  CONFIG = _CONFIG




from datetime import timedelta, datetime
from naeural_core.utils.basic_anomaly_model import BasicAnomalyModel
import numpy as np
DAYS_OF_WEEK = [
  0, #  'MONDAY',
  1, #'TUESDAY',
  2, #'WEDNESDAY',
  3, #'THURSDAY',
  4, #'FRIDAY',
  5, #'SATURDAY',
  6, #'SUNDAY'
]

HOURS = list(range(24))
DAYS_OF_MONTH = list(range(1,32))
EXPECTED_VALUE_INTERVAL_MULTIPLIER = 1.6

class MultiPeriodAnomalyModel:
  def __init__(self, anom_prc, log, timers_section=None, data_validation_callback=None):
    """
    :param anom_prc: float -> prc threshold that an event is considered an anomaly
    :param log: Logger object
    :param timers_section: Needed for logger.P
    :param data_validation_callback: callback used to validate data
    """
    self.anom_prc = anom_prc
    self.models = {
      'HOURLY_DAYS_OF_WEEK': {},
      'DAYS_OF_WEEK': {},
      'DAYS_OF_MONTH': {}
    }
    self.trained = False
    self.data_validation_callback = data_validation_callback
    self.log = log

    self._timers_section = timers_section
    return


  def train(self, np_data_hourly, np_data_daily, dates2id_hourly, dates2ids_daily, filter_start_date=None, new_dates_hourly=None, new_dates_daily=None):
    """
    Trains all the anomaly models where we have new data (eg. if we have only new datapoints for friday, we will only
      train the friday models). Filtering is done by filtering items in the `dates2id` dict

    :param np_data_hourly: Whole hourly data history
    :param np_data_daily: Whole daily data history
    :param dates2id_hourly: Dict mapping datetime to ids - for hourly data
    :param dates2ids_daily: Dict mapping datetime to ids - for daily data
    :param filter_start_date: A start date used to filter the data
    :param new_dates_hourly: Numpy array containing new hourly data (not previously trained on)
    :param new_dates_daily: Numpy array containing new daily data (not previously trained on)
    :return:
    """

    if filter_start_date:
      # Setup the start date filter
      dates2id_hourly = {k:v for k,v in dates2id_hourly.items() if k > filter_start_date}
      dates2ids_daily = {k:v for k,v in dates2ids_daily.items() if k > filter_start_date}
    #endif

    if new_dates_hourly is not None:
      unique_dow_hour = {x: [] for x in np.unique([y.day_of_week for y in new_dates_hourly])}
      for dow in unique_dow_hour.keys():
        unique_dow_hour[dow] = np.unique([x.hour for x in new_dates_hourly if x.day_of_week == dow]).tolist()
      self._train_hourly_days_of_week(np_data_hourly, dates2id_hourly, unique_dow_hour)

    if new_dates_daily is not None:
      unique_dow = np.unique([x.day_of_week for x in new_dates_daily]).tolist()
      self._train_days_of_week(np_data_daily, dates2ids_daily, unique_dow)

      unique_dom = np.unique([x.day for x in new_dates_daily]).tolist()
      self._train_days_of_month(np_data_daily, dates2ids_daily, unique_dom)
    self.trained = True
    return

  def _train_hourly_days_of_week(self, np_data_hourly, dates2id, unique_dow_hour):
    self.log.start_timer('train_hourly_dow', section=self._timers_section)
    for dow in unique_dow_hour.keys():
      filtered_dates2id = {
        k:v for k,v in dates2id.items()
        if k.day_of_week == dow
      }
      if dow not in self.models['HOURLY_DAYS_OF_WEEK']:
        self.models['HOURLY_DAYS_OF_WEEK'][dow] = {}
      # endif
      for hour in unique_dow_hour[dow]:
        self.log.start_timer('train_hourly_dow_single', section=self._timers_section)

        self.log.start_timer('train_hourly_dow_filter', section=self._timers_section)
        ts_indices = [
          v for k,v in filtered_dates2id.items()
          if k.hour == hour
        ]
        ts = np.expand_dims(np_data_hourly[ts_indices], -1)
        self.log.end_timer_no_skip('train_hourly_dow_filter', section=self._timers_section)

        self.log.start_timer('train_hourly_dow_fit', section=self._timers_section)

        self.models['HOURLY_DAYS_OF_WEEK'][dow][hour] = BasicAnomalyModel(
          data_validation_callback=self.data_validation_callback
        )
        self.models['HOURLY_DAYS_OF_WEEK'][dow][hour].fit(
          x_train=ts,
          prc=self.anom_prc
        )
        self.log.end_timer_no_skip('train_hourly_dow_fit', section=self._timers_section)

        self.log.end_timer_no_skip('train_hourly_dow_single', section=self._timers_section)

      #endfor hours
    #endfor dow
    self.log.end_timer_no_skip('train_hourly_dow', section=self._timers_section)
    return

  def _train_days_of_week(self, np_data_daily, dates2id, unique_dow):
    self.log.start_timer('train_dow', section=self._timers_section)

    for dow in unique_dow:
      self.log.start_timer('train_dow_single', section=self._timers_section)

      self.log.start_timer('train_dow_filter', section=self._timers_section)
      ts_indices = [
        v for k, v in dates2id.items()
        if k.day_of_week == dow
      ]
      ts = np.expand_dims(np_data_daily[ts_indices], -1)
      self.log.end_timer_no_skip('train_dow_filter', section=self._timers_section)
      self.log.start_timer('train_dow_fit', section=self._timers_section)

      self.models['DAYS_OF_WEEK'][dow] = BasicAnomalyModel(
        data_validation_callback=self.data_validation_callback
      )
      self.models['DAYS_OF_WEEK'][dow].fit(
          x_train=ts,
          prc=self.anom_prc
      )
      self.log.end_timer_no_skip('train_dow_fit', section=self._timers_section)

      self.log.end_timer_no_skip('train_dow_single', section=self._timers_section)

    #endfor
    self.log.end_timer_no_skip('train_dow', section=self._timers_section)

    return

  def _train_days_of_month(self, np_data_daily, dates2id, unique_dom):
    self.log.start_timer('train_dom', section=self._timers_section)

    for dom in unique_dom:
      self.log.start_timer('train_dom_single', section=self._timers_section)
      self.log.start_timer('train_dom_filter', section=self._timers_section)

      ts_indices = [
        v for k, v in dates2id.items()
        if k.day == (dom + 1)
      ]
      ts = np.expand_dims(np_data_daily[ts_indices], -1)
      self.log.end_timer_no_skip('train_dom_filter', section=self._timers_section)

      self.log.start_timer('train_dom_fit', section=self._timers_section)

      self.models['DAYS_OF_MONTH'][dom] = BasicAnomalyModel(
        data_validation_callback=self.data_validation_callback
      )
      self.models['DAYS_OF_MONTH'][dom].fit(
          x_train=ts,
          prc=self.anom_prc
      )
      self.log.end_timer_no_skip('train_dom_fit', section=self._timers_section)

      self.log.end_timer_no_skip('train_dom_single', section=self._timers_section)

    # endfor
    self.log.end_timer_no_skip('train_dom', section=self._timers_section)

    return

  def predict(self, np_data_hourly=None, np_data_daily=None, dates2ids_hourly=None, dates2ids_daily=None, series_info={}, first_event_day=None, skip_top_prc=None):
    """

    :param np_data_hourly: hourly data to be predicted
    :param np_data_daily: daily data to be predicted
    :param dates2ids_hourly: datetime to ids dict for hourly data
    :param dates2ids_daily: datetime to ids dict for daily data
    :param series_info: metadata to be appended to the anomalies
    :param first_event_day: datetime of the first event in the timeseries (to skip the 0s)
    :param skip_top_prc: remove the first percentage of elements (to skip stuff like configurations etc)

    :return: list(anomalies)
    """
    if not self.trained:
      return None
    anomalies = []

    if first_event_day:
      filter_start_date = first_event_day
      if skip_top_prc is not None:
        no_days = (datetime.now() - filter_start_date).days
        filter_start_date = filter_start_date + timedelta(days=int(no_days * skip_top_prc / 100))
    else:
      filter_start_date = None
    #endif

    if np_data_hourly is not None:
      if filter_start_date:
        dates2ids_hourly = {k: v for k, v in dates2ids_hourly.items() if k >= filter_start_date}
      #endif

      anomalies += self._predict_hourly_days_of_week(np_data_hourly, dates2ids_hourly, first_event_day=first_event_day, series_info=series_info)
    #endif

    if np_data_daily is not None:
      if filter_start_date:
        dates2ids_daily = {k: v for k, v in dates2ids_daily.items() if k >= filter_start_date}
      #endif

      anomalies += self._predict_days_of_week(np_data_daily, dates2ids_daily, first_event_day=first_event_day, series_info=series_info)
      anomalies += self._predict_days_of_month(np_data_daily, dates2ids_daily, first_event_day=first_event_day, series_info=series_info)
    #endif

    return anomalies

  def _predict_hourly_days_of_week(self, np_data_hourly, dates2id, first_event_day, series_info={}):
    anomalies = []

    for dow in DAYS_OF_WEEK:
      filtered_dates2id = {
        k: v for k, v in dates2id.items()
        if k.day_of_week == dow
      }
      if len(filtered_dates2id) == 0:
        continue

      for hour in HOURS:
        ts_indices = [
          v for k, v in filtered_dates2id.items()
          if k.hour == hour
        ]
        if len(ts_indices) == 0:
          continue

        ts = np.expand_dims(np_data_hourly[ts_indices], -1)

        preds = self.models['HOURLY_DAYS_OF_WEEK'][dow][hour].predict(
          x_test=ts
        )

        if any(preds):
          for i in range(preds.sum()):
            anomaly_date = {v: k for k, v in dates2id.items()}[np.array(ts_indices)[preds[:, 0]][i]]

            anomaly = {
              'anomaly_type': 'HOURLY_WEEKDAY_ANOMALY_W{}_H{}'.format(dow, hour),
              'expected_values': (
                max(np.round(self.models['HOURLY_DAYS_OF_WEEK'][dow][hour]._mean[0] - EXPECTED_VALUE_INTERVAL_MULTIPLIER * self.models['HOURLY_DAYS_OF_WEEK'][dow][hour]._std[0], decimals=1), 0),
                np.round(self.models['HOURLY_DAYS_OF_WEEK'][dow][hour]._mean[0] + EXPECTED_VALUE_INTERVAL_MULTIPLIER * self.models['HOURLY_DAYS_OF_WEEK'][dow][hour]._std[0], decimals=1)
              ),
              'anomaly_value': np_data_hourly[dates2id[anomaly_date]],
              'anomaly_date': datetime.strftime(anomaly_date, '%Y-%m-%d %H:%M'),
             'DAYS_FROM_START_DATE': (anomaly_date - first_event_day).days,

              **series_info
            }
            anomalies.append(anomaly)
          # endfor
        #endif
      # endfor hours
    # endfor dow
    return anomalies

  def _predict_days_of_week(self, np_data_daily, dates2id, first_event_day, series_info={}):
    anomalies = []

    for dow in DAYS_OF_WEEK:
      ts_indices = [
        v for k, v in dates2id.items()
        if k.day_of_week == dow
      ]
      ts = np.expand_dims(np_data_daily[ts_indices], -1)

      # Get and save results
      # TODO(S): investigate why does this throw error, key error 0
      preds = self.models['DAYS_OF_WEEK'][dow].predict(
        x_test=ts
      )

      if any(preds):
        for i in range(preds.sum()):
          # Preds is [True, False, False, ...]
          anomaly_date_selector = np.array(ts_indices)[preds[:,0]][i]
          anomaly_date = {
            v:k for k,v in dates2id.items()
          }[anomaly_date_selector]

          anomaly = {
            'anomaly_type': 'WEEKDAY_ANOMALY_{}'.format(dow),
            'expected_values': (
              max(np.round(self.models['DAYS_OF_WEEK'][dow]._mean[0] - EXPECTED_VALUE_INTERVAL_MULTIPLIER * self.models['DAYS_OF_WEEK'][dow]._std[0], decimals=1), 0),
              np.round(self.models['DAYS_OF_WEEK'][dow]._mean[0] + EXPECTED_VALUE_INTERVAL_MULTIPLIER * self.models['DAYS_OF_WEEK'][dow]._std[0], decimals=1)
            ),
            'anomaly_value': np_data_daily[dates2id[anomaly_date]], #ts[np.arange(preds.shape[0])[preds[:,0]][i]][0],
            'anomaly_date': datetime.strftime(anomaly_date, '%Y-%m-%d'),
            'DAYS_FROM_START_DATE': (anomaly_date - first_event_day).days,
            **series_info
          }
          anomalies.append(anomaly)
        #endfor
      #endif
    #endfor
    return anomalies

  def _predict_days_of_month(self, np_data_daily, dates2id, first_event_day, series_info={}):
    anomalies = []

    for dom in DAYS_OF_MONTH:
      ts_indices = [
        v for k, v in dates2id.items()
        if k.day == dom
      ]
      ts = np.expand_dims(np_data_daily[ts_indices], -1)
      # Get and save results
      if ts.shape[0] > 0:
        preds = self.models['DAYS_OF_MONTH'][dom].predict(
          x_test=ts,
        )
        if any(preds):
          for i in range(preds.sum()):
            anomaly_date = {v: k for k, v in dates2id.items()}[np.array(ts_indices)[preds[:, 0]][i]]
            anomaly = {
              'anomaly_type': 'MONTHDAY_ANOMALY_{}'.format(dom),
              'expected_values': (
                max(np.round(self.models['DAYS_OF_MONTH'][dom]._mean[0] - EXPECTED_VALUE_INTERVAL_MULTIPLIER * self.models['DAYS_OF_MONTH'][dom]._std[0], decimals=1), 0),
                np.round(self.models['DAYS_OF_MONTH'][dom]._mean[0] + EXPECTED_VALUE_INTERVAL_MULTIPLIER * self.models['DAYS_OF_MONTH'][dom]._std[0], decimals=1)
              ),
              'anomaly_value': np_data_daily[dates2id[anomaly_date]],
              'anomaly_date': datetime.strftime(anomaly_date, '%Y-%m-%d'),
              'DAYS_FROM_START_DATE': (anomaly_date - first_event_day).days,
              **series_info
            }
            anomalies.append(anomaly)
          #endfor range(preds.sum())
        #endif any(preds)
      #endif ts.shape[0]
    #endfor
    return anomalies



if __name__ == '__main__':
  from naeural_core import Logger
  # from datetime import datetime
  from naeural_core.utils.predictive_analytics.timeseries.numpyize_time_series import TimeseriesDataMaster

  log = Logger(lib_name='TST', base_folder='dropbox', app_folder='_local_data/_product_dynamics/_goc_experiments', TF_KERAS=False)
  data = log.load_dataframe(fn='goc_data.csv', timestamps=['DT_hour']).drop(columns=['Unnamed: 0'])

  initial_data = data[data.DT_hour < datetime(year=2022, month=2, day=1)]
  delta_data = data[data.DT_hour > datetime(year=2022, month=2, day=1)]

  eng = TimeseriesDataMaster(
    log=log, prefix='to_be_changed',
    aggregation_fields=['ID_SubEchipament', 'CheieEveniment'],
    time_field='DT_hour',
    target_field='event_count',
  )  ### TODO: parametrize

  # TODO: if is not None
  ds_master = eng.create_dataset()
  ds_master.setup(
    column_names=['event_count'],
    freq='d',  # set to 'h'
    df=initial_data
  )
  dct_datasets_master, _, _ = ds_master()
  np_master = dct_datasets_master['event_count']

  anomaly_model = MultiPeriodAnomalyModel(
    anom_prc=0.02
  )

  anomaly_model.train(
    np_master[3368], ds_master.dates2id
  )

  ds_delta = eng.create_dataset()
  ds_delta.setup(
    column_names=['event_count'],
    freq='h',  # set to 'h'
    df=delta_data
  )
  dct_datasets_delta, _, _ = ds_delta()
  np_delta = dct_datasets_delta['event_count']

  anomalies = anomaly_model.predict(
    np_delta[3368], ds_delta.dates2id
  )







import threading
import ngrok
import asyncio
__VER__ = '0.0.0.0'


class _NgrokMixinPlugin(object):
  class NgrokCT:
    NG_TOKEN = 'EE_NGROK_AUTH_TOKEN'
    # HTTP_GET = 'get'
    # HTTP_PUT = 'put'
    # HTTP_POST = 'post'

  """
  A plugin which exposes all of its methods marked with @endpoint through
  fastapi as http endpoints, and further tunnels traffic to this interface
  via ngrok.

  The @endpoint methods can be triggered via http requests on the web server
  and will be processed as part of the business plugin loop.
  """

  def _reset_ngrok(self):
    self.ngrok_listener = None
    self.ngrok_started = False
    self.ngrok_initiated = False
    return

  @property
  def app_url(self):
    return None if self.ngrok_listener is None else self.ngrok_listener.url()

  def get_setup_commands(self):
    try:
      super_setup_commands = super(_NgrokMixinPlugin, self).get_setup_commands()
    except AttributeError:
      super_setup_commands = []
    if self.cfg_ngrok_use_api:
      # In this case the authentification will be made through the api in the actual code,
      # instead of the command line.
      return super_setup_commands
    # endif ngrok api used

    if self.cfg_ngrok_enabled:
      return [self.__get_ngrok_auth_command()] + super_setup_commands
    else:
      return super_setup_commands

  def report_missing_authtoken(self):
    msg = "Ngrok token not found. Please set the environment variable `EE_NGROK_AUTH_TOKEN`"
    # Maybe have notif_code in the future
    self.P(msg, color='r')
    self._create_notification(
      msg=msg,
    )
    return

  def maybe_init_ngrok(self):
    if self.cfg_ngrok_use_api and not self.ngrok_initiated:
      self.ngrok_initiated = True
      ng_token = self.__get_ng_token()
      if ng_token is None:
        self.report_missing_authtoken()
      else:
        ngrok.set_auth_token(ng_token)
        self.P(f"Ngrok initiated for {self.unique_identification}.")
      # endif ng_token present
    # endif ngrok api used
    return

  def get_ngrok_tunnel_kwargs(self):
    """
    TODO:
      - in the case of container in container we will need to add `addr` parameter to the
        tunnel_kwargs as we might have a local network within the Edge Node.
    """
    # Make the ngrok tunnel kwargs
    tunnel_kwargs = {}
    valid = True
    if self.cfg_ngrok_edge_label in [None, '']:
      self.P("WARNING: ngrok edge label is not set. Please make sure this is the intended behavior.", color='r')
    # endif edge label
    if self.cfg_ngrok_edge_label is not None:
      # In case of using edge label, the domain is not needed and the protocol is "labeled".
      tunnel_kwargs['labels'] = f'edge:{self.cfg_ngrok_edge_label}'
      tunnel_kwargs['proto'] = "labeled"
    # endif edge label
    elif self.cfg_ngrok_domain is not None:
      # In case of using domain, the domain is needed and the protocol is "http"(the default value).
      tunnel_kwargs['domain'] = self.cfg_ngrok_domain
    # endif domain
    # Specify the address and the authtoken
    tunnel_kwargs['addr'] = self.port
    ng_token = self.__get_ng_token()
    if ng_token is None:
      valid = False
      self.report_missing_authtoken()
    tunnel_kwargs['authtoken'] = ng_token
    return tunnel_kwargs, valid


  async def __maybe_async_stop_ngrok(self):
    try:
      self.P(f"Ngrok stopping...")
      self.ngrok_listener.close()
      self.ngrok_started = False
      self.P(f"Ngrok stopped.")
    except Exception as exc:
      self.P(f"Error stopping ngrok: {exc}", color='r')
    return

  
  def maybe_stop_ngrok(self):
    if self.ngrok_started and self.ngrok_listener is not None:
      self.P(f"Closing Ngrok listener...")

      threading.Thread(target=lambda: asyncio.run(self.__maybe_async_stop_ngrok()), daemon=True).start()
    return


  def maybe_start_ngrok(self):
    """
    
    TODO: 
      - if no edge/domain is specified, the api should generate a url and return it while
        persisting the url in the instance local cache for future use. When the instance is 
        restarted, the same url should be used.
    """
    # Maybe make this asynchronous?
    if self.cfg_ngrok_use_api and not self.ngrok_started:
      self.P(f"Ngrok starting for {self.unique_identification}...")
      tunnel_kwargs, valid = self.get_ngrok_tunnel_kwargs()
      if valid:
        if self.cfg_debug_web_app:
          self.P(f'Ngrok tunnel kwargs: {tunnel_kwargs}')
        self.ngrok_listener = ngrok.forward(**tunnel_kwargs)
        if self.app_url is not None:
          self.P(f"Ngrok started on URL `{self.app_url}` ({self._signature}).")
        else:
          edge = tunnel_kwargs.get('labels')
          domain = tunnel_kwargs.get('domain')
          str_tunnel = f"edge `{edge}`" if edge is not None else f"domain `{domain}`"
          self.P(f"Ngrok started on {str_tunnel} ({self._signature}).")
        # endif app_url
        self.ngrok_started = True
      # endif valid tunnel kwargs
    # endif ngrok api used and not started
    return

  def get_start_commands(self):
    try:
      super_start_commands = super(_NgrokMixinPlugin, self).get_start_commands()
    except AttributeError:
      super_start_commands = []
    if self.cfg_ngrok_use_api:
      # In case of using the ngrok api, the tunnel will be started through the api
      return super_start_commands
    # endif ngrok api used

    if self.cfg_ngrok_enabled:
      return [self.__get_ngrok_start_command()] + super_start_commands
    else:
      return super_start_commands

  def __get_ng_token(self):
    # TODO: At the moment multiple user auth tokens will not work on the same node
    #  if `NGROK_USE_API` is set to False.
    #  For the same node to use more than one auth token it needs to do so in separated
    #  processes. For the moment this is done only if the ngrok is started through
    #  CLI commands.
    configured_ng_token = self.cfg_ngrok_auth_token
    environment_ng_token = self.os_environ.get(_NgrokMixinPlugin.NgrokCT.NG_TOKEN, None)
    if self.cfg_debug_web_app:
      self.P(f"Configured token: {configured_ng_token}, Environment token: {environment_ng_token}")
    return configured_ng_token if configured_ng_token is not None else environment_ng_token

  def __get_ngrok_auth_command(self):
    return f"ngrok authtoken {self.__get_ng_token()}"

  def __get_ngrok_start_command(self):
    if self.cfg_ngrok_edge_label is not None:
      return f"ngrok tunnel {self.port} --label edge={self.cfg_ngrok_edge_label}"
    elif self.cfg_ngrok_domain is not None:
      return f"ngrok http {self.port} --domain={self.cfg_ngrok_domain}"
    else:
      raise RuntimeError("No domain/edge specified. Please check your configuration.")
    # endif



from naeural_core.business.base import BasePluginExecutor as BaseClass
from naeural_core.business.mixins_libs.ai4e_mixin import _Ai4eMixin, AI4E_MIXIN_CONFIG
from os import walk


_CONFIG = {
  **BaseClass.CONFIG,
  **AI4E_MIXIN_CONFIG,
  # TODO: add job initiator here and in crop_data
  'ALLOW_EMPTY_INPUTS': True,
  'RUN_WITHOUT_IMAGE': True,
  "FULL_DEBUG_LOG": True,
  "CLASSES": None,
  'TRAIN_SIZE': 0.8,
  'MIN_DECISION_PRC': 0.5,

  "LABEL_BACKUP_PERIOD": 10,
  'CLOUD_PATH': 'DATASETS/',
  'LABEL_UNTIL': None,

  'VALIDATION_RULES': {
    **BaseClass.CONFIG['VALIDATION_RULES'],
    **AI4E_MIXIN_CONFIG['VALIDATION_RULES'],
  },
}


class Ai4eLabelDataPlugin(BaseClass, _Ai4eMixin):
  _CONFIG = _CONFIG

  def debug_log(self, msg, **kwargs):
    if self.cfg_full_debug_log:
      self.P(msg, **kwargs)
    return

  def on_init(self):
    super(Ai4eLabelDataPlugin, self).on_init()
    self.accepting_votes = False
    self.started_label_finish = False
    self.label_updates_count = 0
    self.raw_dataset_path = None
    self.raw_dataset_rel_path = None
    self.final_dataset_rel_path = self.os_path.join('final_datasets', self.cfg_objective_name)
    self.final_dataset_full_path = self.os_path.join(self.get_output_folder(), self.final_dataset_rel_path)

    self.filenames = set()
    # All the votes in format
    # {'filename1': {'class1': nr_votes1}}
    self.filename_votes = {}
    # All the votes in format
    # {'worker_id': {'filename1': {'class1': nr_votes1}}}
    self.worker_votes = {}
    # All the decisions in format
    # {'filename1': 'class1'}
    # The above means that for filename1 the class1 was decided by the majority of votes
    self.filename_decisions = {}
    # All the stats for the decisions
    # {'class1': count1}
    # The above means that class1 was decided for count1 files
    self.decision_stats = self.default_classes_stats()
    self.done_labeling = False
    self.publish_status = 0
    self.train_subdirs = {}
    self.train_subdir_stats = {
      'train': 0,
      'dev': 0,
    }
    self.maybe_load_previous_state()
    return

  def maybe_load_previous_state(self):
    self.debug_log('Loading previous state')
    obj = self.persistence_serialization_load()
    if obj is None:
      return
    self.debug_log(f'Loaded previous state: {obj}')
    self.accepting_votes = obj.get('accepting_votes', self.accepting_votes)
    self.started_label_finish = obj.get('started_label_finish', self.started_label_finish)
    self.label_updates_count = obj.get('label_updates_count', self.label_updates_count)
    self.raw_dataset_path = obj.get('raw_dataset_path', self.raw_dataset_path)
    self.raw_dataset_rel_path = obj.get('raw_dataset_rel_path', self.raw_dataset_rel_path)

    self.filenames = obj.get('filenames', self.filenames)
    self.filename_votes = obj.get('filename_votes', self.filename_votes)
    self.worker_votes = obj.get('worker_votes', self.worker_votes)
    self.filename_decisions = obj.get('filename_decisions', self.filename_decisions)
    self.decision_stats = obj.get('decision_stats', self.decision_stats)
    self.done_labeling = obj.get('done_labeling', self.done_labeling)
    self.publish_status = obj.get('publish_status', self.publish_status)
    self.train_subdirs = obj.get('train_subdirs', self.train_subdirs)
    self.train_subdir_stats = obj.get('train_subdir_stats', self.train_subdir_stats)
    return

  def maybe_save_current_state(self, force=False):
    to_save = force
    if self.label_updates_count != 0 and self.label_updates_count % self.cfg_label_backup_period == 0:
      to_save = True
    if not to_save:
      return
    self.debug_log('Saving current state')
    obj = {
      'accepting_votes': self.accepting_votes,
      'started_label_finish': self.started_label_finish,
      'label_updates_count': self.label_updates_count,
      'raw_dataset_path': self.raw_dataset_path,
      'raw_dataset_rel_path': self.raw_dataset_rel_path,

      'filenames': self.filenames,
      'filename_votes': self.filename_votes,
      'worker_votes': self.worker_votes,
      'filename_decisions': self.filename_decisions,
      'decision_stats': self.decision_stats,
      'done_labeling': self.done_labeling,
      'publish_status': self.publish_status,
      'train_subdirs': self.train_subdirs,
      'train_subdir_stats': self.train_subdir_stats,
    }
    self.persistence_serialization_save(obj)
    return

  def delay_process(self):
    # Raw dataset initialized and still accepting labels
    if not self.started_label_finish:
      return False
    # Raw dataset initialized, no longer accepting labels
    # due to finalizing the label files
    if not self.done_labeling:
      return True
    # Final dataset ready to be published
    return False

  def get_plugin_loop_resolution(self):
    return self.cfg_plugin_loop_resolution if not self.delay_process() else 1 / 30


  def maybe_read_filenames(self, root_path=None):
    if len(self.filenames) == 0 and root_path is not None:
      start_idx = len(root_path) + len(self.os_path.sep)
      # TODO: maybe make diskapi_walk for safety instead of plain walk
      for (dirpath, dirnames, filenames) in walk(root_path):
        rel_path = dirpath[start_idx:]
        for filename in filenames:
          self.filenames.add(self.os_path.join(rel_path, self.os_path.splitext(filename)[0]))
        # endfor filenames
      # endfor walk
      self.raw_dataset_path = root_path
      output_absolute_path = self.os_path.abspath(self.log.get_data_folder())
      root_path_absolute = self.os_path.abspath(root_path)
      self.raw_dataset_rel_path = self.os_path.relpath(root_path_absolute, output_absolute_path)
      self.accepting_votes = True
      self.maybe_save_current_state(force=True)
    # endif no filenames
    return

  def get_present_workers(self):
    # TODO: implement
    return 0

  @property
  def dataset_object_name_final(self):
    return self.os_path.join(self.cfg_cloud_path, self.cfg_objective_name + '.zip')

  def get_job_status(self):
    status = "Publishing"
    if len(self.filenames) > 0:
      status = "Labeling"
    if self.done_labeling:
      status = "Done Labeling"
    if self.publish_status != 0:
      status = "Publishing labels"
    return status

  def get_training_subdir(self):
    """
    Get the training subdirectory based on the configured TRAIN_SIZE.
    Returns
    -------
    str, the training subdirectory
    """
    if (self.train_subdir_stats['train'] * self.train_subdir_stats['dev']) == 0:
      if self.train_subdir_stats['train'] == 0:
        return 'train'
      return 'dev'
    # endif no files in train or dev
    train_size = self.cfg_train_size
    train_size = max(0.0, min(1.0, train_size))
    choice = self.np.random.choice([0, 1], p=[1 - train_size, train_size])
    train_subdir = 'dev' if choice == 0 else 'train'
    return train_subdir

  def default_classes_stats(self):
    return {
      k: 0 for k in self.get_ds_classes().keys()
    }

  def maybe_create_additional_file(self):
    dataset_info_path = self.os_path.join(self.final_dataset_rel_path, 'ADDITIONAL.json')
    if self.os_path.exists(dataset_info_path):
      return
    classes_data = {
      'classes': self.get_ds_classes(),
      'name': self.cfg_objective_name,
    }
    classes_path = self.diskapi_save_json_to_output(
      dct=classes_data,
      filename=dataset_info_path,
    )
    return

  """VOTE SECTION"""
  if True:
    def maybe_register_filename(self, filename: str, votes_dict: dict=None):
      if votes_dict is None:
        votes_dict = self.filename_votes
      if filename not in votes_dict.keys():
        votes_dict[filename] = self.default_classes_stats()
      return

    def maybe_register_worker(self, worker_id):
      if worker_id not in self.worker_votes.keys():
        self.worker_votes[worker_id] = {}
      return self.worker_votes[worker_id]

    def copy_datapoint_to_final_dataset(self, filename):
      raw_path = self.os_path.join(self.raw_dataset_path, f'{filename}.jpg')
      train_subdir = self.get_training_subdir()
      final_path = self.os_path.join(
        self.final_dataset_full_path, train_subdir, f'{filename}.jpg'
      )
      self.debug_log(f'Copying {raw_path} to {final_path}')
      self.train_subdirs[filename] = train_subdir
      self.train_subdir_stats[train_subdir] += 1
      self.diskapi_copy_file(
        src_path=raw_path,
        dst_path=final_path
      )
      return

    def remove_datapoint_from_final_dataset(self, filename):
      train_subdir = self.train_subdirs[filename]
      final_path = self.os_path.join(
        self.final_dataset_full_path, train_subdir, f'{filename}.jpg'
      )
      del self.train_subdirs[filename]
      self.debug_log(f'Removing {final_path}')
      self.train_subdir_stats[train_subdir] -= 1
      self.diskapi_delete_file(final_path)
      return

    def maybe_add_decision(self, filename, decision):
      if filename not in self.filename_decisions.keys():
        # The filename did not have a decided label before
        # Copy it to the final dataset
        self.copy_datapoint_to_final_dataset(filename)
        self.decision_stats[decision] += 1
      else:
        # The filename had a decided label before.
        # Remove it from the statistics in case the decision changes.
        prev_decision = self.filename_decisions[filename]
        if prev_decision != decision:
          self.decision_stats[prev_decision] -= 1
          self.decision_stats[decision] += 1
        # endif decision changed
      # endif filename had a decided label before
      # Add the new decision to the statistics
      self.filename_decisions[filename] = decision
      return

    def maybe_remove_decision(self, filename):
      if filename not in self.filename_decisions.keys():
        # The filename does not have a decided label so nothing to do.
        return
      # endif filename had a decided label
      # The filename had a decided label, so it will be removed from the final dataset.
      self.remove_datapoint_from_final_dataset(filename)
      decision = self.filename_decisions[filename]
      self.decision_stats[decision] -= 1
      del self.filename_decisions[filename]
      return

    def maybe_update_decision(self, filename):
      if filename not in self.filename_votes.keys():
        return
      self.debug_log(f'Updating decision for {filename}')
      votes = self.filename_votes[filename]
      max_votes = max(votes.values())
      total_votes = sum(votes.values())
      # If the minimum percentage is 0.5 the decision is made if the max votes are at least 50% + 1 of the total votes
      is_decided = max_votes >= int(total_votes * self.cfg_min_decision_prc + 1)
      decision = [k for k, v in votes.items() if v == max_votes][0]
      # If the label of the filename is decided
      self.debug_log(f'[{is_decided=}]Decision for {filename} is {decision}')
      if is_decided:
        self.maybe_add_decision(filename, decision)
      else:
        self.maybe_remove_decision(filename)
      # endif is decided
      max_votes_classes = [k for k, v in votes.items() if v == max_votes]
      if len(max_votes_classes) == 1:
        self.filename_decisions[filename] = max_votes_classes[0]
        self.decision_stats[max_votes_classes[0]] += 1
      return

    def count_vote(self, filename, label, worker_id):
      """
      Method for counting a vote for a filename and label and updating the stats.
      Parameters
      ----------
      filename : str, the filename
      label : str, the label
      worker_id : str, the worker id

      Returns
      -------
      bool : True if the vote was counted, False otherwise
      """
      # Filename does not exist in the raw dataset, do nothing
      if filename not in self.filenames:
        return False
      # Label does not exist in the dataset classes, do nothing
      if label not in self.get_ds_classes().keys():
        return False
      # If the worker is seen for the first time, its statistics are initiated
      worker_votes = self.maybe_register_worker(worker_id)
      # If the filename is seen for the first time, its statistics are initiated
      self.maybe_register_filename(filename)
      # If the filename is seen for the first time by this worker, its statistics are initiated
      self.maybe_register_filename(filename, worker_votes)
      self.filename_votes[filename][label] += 1
      worker_votes[filename][label] += 1
      self.maybe_update_decision(filename)
      return True

    def finalize_labels(self):
      if self.done_labeling:
        return
      self.P("Finalizing labels", color='g')
      self.accepting_votes = False
      self.started_label_finish = True
      self.maybe_save_current_state(force=True)
      self.P('Saving the labels according to the majority votes', color='g')
      for filename, decision in self.filename_decisions.items():
        train_subdir = self.train_subdirs.get(filename)
        if train_subdir is None:
          continue
        label_subdir = self.os_path.join(self.final_dataset_rel_path, train_subdir)
        label_path = self.os_path.join(label_subdir, f'{filename}.txt')
        if self.os_path.exists(label_path):
          continue
        self.diskapi_save_file_to_output(
          data=decision,
          filename=f'{filename}.txt',
          subdir=label_subdir
        )
      # endfor filename, decision
      self.done_labeling = True
      self.maybe_save_current_state(force=True)
      return
  """END VOTE SECTION"""

  """COMMAND SECTION"""
  if True:
    def label_until_passed(self):
      label_until = self.cfg_label_until
      if label_until is None:
        return False
      return self.datetime.fromtimestamp(label_until) < self.datetime.now()

    def maybe_handle_datapoint_label(self, datapoint: dict, worker_id: str):
      """
      Method for handling the vote from a worker_id about a datapoint.
      The datapoint has to contain both the filename and the label.
      Parameters
      ----------
      datapoint : dict, the datapoint to label
      worker_id : str, the worker id

      Returns
      -------

      """
      # If the voting is done, no more votes are accepted.
      if not self.accepting_votes or self.label_until_passed():
        return
      # No datapoint is present, nothing is done.
      if datapoint is None:
        return
      datapoint = {
        k.upper(): v for k, v in datapoint.items()
      }
      filename = datapoint.get('FILENAME', None)
      label = datapoint.get('LABEL', None)
      # No filename or label is present, nothing is done.
      if label is None or filename is None:
        return
      counted = self.count_vote(filename, label, worker_id)
      if counted:
        # Increment the label updates count
        self.label_updates_count += 1
        self.maybe_save_current_state()
      return

    def filename_to_path(self, filename):
      return self.os_path.join(self.raw_dataset_path, filename)

    def process_request(self, data, raw=True, filename=None):
      request_id = data.get('REQUEST_ID')
      if filename is None:
        filenames = list(self.filenames) if raw else list(self.filename_votes.keys())
        filename = self.np.random.choice(filenames) if len(filenames) > 0 else None
      # endif filename not provided
      if filename is None:
        response_kwargs = {
          'request_id': request_id,
          'is_status': False,
          'status': 'No data available',
        }
        self.add_payload_by_fields(**response_kwargs)
        return
      img = self.diskapi_load_image(
        folder='data',
        filename=f'{filename}.jpg',
        subdir=self.raw_dataset_rel_path,
      )
      response_kwargs = {
        'request_id': request_id,
        'is_status': False,
        'img': img,
        'sample_filename': filename,
        'file_path': self.filename_to_path(filename)
      }
      if not raw:
        response_kwargs['votes'] = self.filename_votes.get(filename, {})
      self.add_payload_by_fields(**response_kwargs)
      return

    def publish_dataset(self):
      self.P(f'Received request for publishing dataset.')
      self.publish_status = 1
      self.maybe_save_current_state(force=True)
      return

    def on_command(self, data: dict, **kwargs):
      is_sample_datapoint_request = data.get('SAMPLE_DATAPOINT', False)
      is_sample_request = data.get('SAMPLE', False)
      is_filename_request = data.get('FILENAME_REQUEST', False)
      # In case of data request, payload with random filename is sent.
      if is_sample_request:
        self.process_request(data)
        return
      # endif data request
      # In case of sample request, payload with random filename, data and votes are sent
      if is_sample_datapoint_request:
        self.process_request(data, raw=False)
        return
      # endif sample request
      # In case of filename request, payload with data and votes
      if is_filename_request:
        self.process_request(data, raw=False, filename=data.get('FILENAME'))
      # endif filename request
      self.maybe_create_additional_file()
      datapoint = data.get('DATAPOINT', None)
      worker_id = data.get('WORKER_ID', None)
      self.maybe_handle_datapoint_label(datapoint, worker_id)

      finish_labeling = data.get('FINISH_LABELING', False)
      if finish_labeling:
        self.finalize_labels()
      # endif finish_labeling
      publish_dataset = data.get('PUBLISH', False)
      if publish_dataset:
        self.publish_dataset()
      return
  """END COMMAND SECTION"""

  def get_status_payload_additional_kwargs(self, **kwargs):
    return {
      'worker_count': self.get_present_workers(),
      'total_files_voted': len(self.filename_votes.keys()),
      'total_files_decided': len(self.filename_decisions.keys()),
      'decided_stats': self.decision_stats,
    }

  def maybe_publish_final(self):
    if self.publish_status == 1:
      if self.done_labeling:
        config = {
          'NAME': f'upload_final_{self.get_instance_id()}',
          'TYPE': 'VOID',
          'PLUGINS': [{
            'SIGNATURE': 'minio_upload_dataset',
            'INSTANCES': [{
              'INSTANCE_ID': self.get_instance_id(),
              'DATASET_OBJECT_NAME': self.dataset_object_name_final,
              'DATASET_LOCAL_PATH': self.final_dataset_full_path,
            }]
          }]
        }
        self.cmdapi_start_pipeline(config=config)
        self.publish_status = 2
      else:
        self.P('Labeling not finished yet, cannot publish the dataset.')
    return

  def maybe_finalize_labels(self):
    if self.done_labeling:
      return
    if self.label_until_passed():
      self.finalize_labels()
    return

  def _process(self):
    self.maybe_report_status()
    ds_status = self.dataapi_struct_data()
    if ds_status is None:
      return
    is_dataset_ready = ds_status.get('dataset_ready', False)
    if not is_dataset_ready:
      return

    self.maybe_finalize_labels()
    self.maybe_publish_final()
    self.maybe_read_filenames(root_path=ds_status.get('dataset_path'))
    return




import multiprocessing as mp

from naeural_core import Logger

from naeural_core.serving.serving_manager import ServingManager

if __name__ == '__main__':
  mp.set_start_method('spawn')
  log = Logger('MPTF', base_folder='.', app_folder='_local_cache', TF_KERAS=False)

  MODEL_NAME =( 'th_y5s6s', 'default')
  NR_TESTS = 10

  # defining script constants
  SERVING_MANAGER = ServingManager(
    shmem={},
    log=log,
    prefix_log='[TFMGR]'
  )
  for i in range(1, NR_TESTS + 1):
    if i == 2:
      log.P("DEBUG!")
    SERVING_MANAGER.start_server(
      server_name=MODEL_NAME,
      inprocess=False,
      upstream_config={
        'USE_AMP': True, 'USE_FP16': False, 'GPU_PREPROCESS': True, 'DEFAULT_DEVICE': 'cuda:0',
       'MAX_BATCH_FIRST_STAGE': 1}
    )

    SERVING_MANAGER.stop_server(server_name=MODEL_NAME)



from naeural_core.local_libraries.nn import utils as nn_utils
from naeural_core.local_libraries.nn.th import utils as th_utils
from naeural_core.local_libraries.nn.th.image_dataset_stage_preprocesser import preprocess_to_str, str_to_preprocess

from naeural_core.utils.plugins_base.plugin_base_utils import _UtilsBaseMixin

class _InferenceUtilsMixin(_UtilsBaseMixin):
  @property
  def nn_utils(self):
    """
    Proxy method for using the nn_utils module
    Returns
    -------
    nn_utils - module, module containing utility functions for neural networks
    """
    return nn_utils

  @property
  def th_utils(self):
    """
    Proxy method for using the th_utils module
    Returns
    -------
    th_utils - module, module containing utility functions for torch
    """
    return th_utils

  def str_to_preprocess(self, preprocess_str):
    return str_to_preprocess(preprocess_str)

  def preprocess_to_str(self, preprocess_obj):
    return preprocess_to_str(preprocess_obj)

  def __init__(self):
    super(_InferenceUtilsMixin, self).__init__()
    return
  
  
  def th_resize_with_pad(self,img, h, w, 
                         device=None, 
                         pad_post_normalize=False,
                         normalize=True,
                         sub_val=0, div_val=255,
                         half=False,
                         return_original=False,
                         normalize_callback=None,
                         fill_value=None,
                         return_pad=False,
                         **kwargs
                         ):
    return th_utils.th_resize_with_pad(
      img, h, w, 
      device=device, 
      pad_post_normalize=pad_post_normalize,
      normalize=normalize,
      sub_val=sub_val, div_val=div_val,
      half=half,
      return_original=return_original,
      normalize_callback=normalize_callback,
      fill_value=fill_value,
      return_pad=return_pad,
      **kwargs
    )
  



def K_rec(y_true, y_pred, threshold=0.5):
  import tensorflow as tf
  K = tf.keras.backend
  y_pred = tf.cast(y_pred >= threshold, dtype=tf.int32)
  y_true = tf.cast(y_true, dtype=tf.int32)
  recall = K.sum(y_true * y_pred) / K.sum(y_true)
  return recall

def K_MAPE_TS_LAST(y_true, y_pred):
  import tensorflow as tf
  K = tf.keras.backend
  tf_res = K.abs(y_true[:, -1, :] - y_pred[:, -1, :])
  tf_yt = K.clip(K.abs(y_true[:, -1, :]), K.epsilon(), None)
  tf_mape = tf_res / tf_yt
  mape_series = K.clip(tf_mape, K.epsilon(), 2)
  return K.mean(mape_series)

def K_pre(y_true, y_pred, threshold=0.5, ):
  import tensorflow as tf
  K = tf.keras.backend
  y_pred = tf.cast(y_pred >= threshold, dtype=tf.int32)
  y_true = tf.cast(y_true, dtype=tf.int32)
  precision = K.sum(y_true * y_pred) / K.sum(y_pred)
  return precision

def K_f2(y_true, y_pred):
  precision = K_pre(y_true, y_pred)
  recall = K_rec(y_true, y_pred)
  return 5 * (precision * recall) / (4 * precision + recall)

def K_f1(y_true, y_pred):
  precision = K_pre(y_true, y_pred)
  recall = K_rec(y_true, y_pred)
  return 2 * (precision * recall) / (precision + recall)

def get_K_clf_metrics():
  return ['accuracy', K_rec, K_pre, K_f2]

def get_K_metrics():
  metrics = get_K_clf_custom_dict()
  return metrics

def get_K_clf_custom_dict():
  return {'K_rec': K_rec, 'K_pre': K_pre, 'K_f2': K_f2}

def K_r2_1D(y_true, y_pred):
  import tensorflow as tf
  K = tf.keras.backend
  SS_res = K.sum(K.square(y_true - y_pred))
  SS_tot = K.sum(K.square(y_true - K.mean(y_true)))
  return 1 - SS_res / (SS_tot + K.epsilon())

def K_r2_TS(y_true, y_pred, return_series=False):
  import tensorflow as tf
  K = tf.keras.backend
  tf_SS_res = K.sum(K.square(y_true - y_pred), axis=1)
  tf_SS_tot = K.sum(K.square(y_true - K.mean(y_true, axis=1,
                                             keepdims=True)),
                    axis=1)
  tf_SS_tot = K.clip(tf_SS_tot, K.epsilon(), None)
  tf_all_R2 = 1 - tf_SS_res / tf_SS_tot
  tf_all_R2 = K.clip(tf_all_R2, -1.5, None)
  if return_series:
    return tf_all_R2
  else:
    return K.mean(tf_all_R2)

def K_MAD(y_true, y_pred, relu=True, std=None, mean=None):
  import tensorflow as tf
  K = tf.keras.backend
  if std is not None and mean is not None:
    y_true = y_true * std + mean
    y_pred = y_pred * std + mean

  if relu:
    y_pred = K.relu(y_pred)

  y_true_s = K.sum(y_true, axis=1)
  y_pred_s = K.sum(y_pred, axis=1)

  y_true_s = K.reshape(y_true_s, shape=[-1, ])
  y_pred_s = K.reshape(y_pred_s, shape=[-1, ])

  tf_zeros = K.zeros(shape=K.shape(y_pred_s))
  not_zero_indices = K.not_equal(y_pred_s, tf_zeros)

  y_true_s = tf.boolean_mask(y_true_s, not_zero_indices)
  y_pred_s = tf.boolean_mask(y_pred_s, not_zero_indices)

  mad = K.abs(y_true_s - y_pred_s) / y_pred_s
  return K.mean(mad)

def K_MAD_TS(y_true, y_pred, relu=True, ERR_MAX=2,
             std=None, mean=None):
  import tensorflow as tf
  K = tf.keras.backend

  if std is not None and mean is not None:
    y_true = y_true * std + mean
    y_pred = y_pred * std + mean

  if relu:
    y_pred = K.relu(y_pred)

  y_true_s = K.sum(y_true, axis=1)
  y_pred_s = K.sum(y_pred, axis=1)

  y_pred_s = K.clip(y_pred_s, K.epsilon(), None)

  tf_mad_pre = K.abs(y_true_s - y_pred_s) / y_pred_s
  tf_mad = K.clip(tf_mad_pre, 0, ERR_MAX)
  return K.mean(tf_mad)


def generate_K_MAD(mean, std, relu=True):
  from functools import partial
  generated = partial(K_MAD, std=std, mean=mean, relu=relu)
  generated.__name__ = K_MAD.__name__
  return generated

def generate_K_MAD_TS(mean, std, relu=True, ERR_MAX=2):
  from functools import partial
  generated = partial(K_MAD_TS, std=std, mean=mean,
                      relu=relu, ERR_MAX=ERR_MAX)
  generated.__name__ = K_MAD_TS.__name__
  return generated
